@@ -1,10 +1,14 @@
-// ANCHOR: here
+use std::sync::mpsc;
 use std::thread;
 
 pub struct ThreadPool {
-    threads: Vec<thread::JoinHandle<()>>,
+    workers: Vec<Worker>,
+    sender: mpsc::Sender<Job>,
 }
 
+struct Job;
+
+// ANCHOR: here
 impl ThreadPool {
     // --snip--
     // ANCHOR_END: here
@@ -19,15 +23,16 @@ impl ThreadPool {
     pub fn new(size: usize) -> ThreadPool {
         assert!(size > 0);
 
-        let mut threads = Vec::with_capacity(size);
+        let (sender, receiver) = mpsc::channel();
 
-        for _ in 0..size {
-            // create some threads and store them in the vector
+        let mut workers = Vec::with_capacity(size);
+
+        for id in 0..size {
+            workers.push(Worker::new(id, receiver));
         }
 
-        ThreadPool { threads }
+        ThreadPool { workers, sender }
     }
-
     // --snip--
     // ANCHOR_END: here
 
@@ -38,4 +43,24 @@ impl ThreadPool {
     }
     // ANCHOR: here
 }
+
+// --snip--
+
+// ANCHOR_END: here
+
+struct Worker {
+    id: usize,
+    thread: thread::JoinHandle<()>,
+}
+
+// ANCHOR: here
+impl Worker {
+    fn new(id: usize, receiver: mpsc::Receiver<Job>) -> Worker {
+        let thread = thread::spawn(|| {
+            receiver;
+        });
+
+        Worker { id, thread }
+    }
+}
 // ANCHOR_END: here
