@@ -1,21 +1,17 @@
-use rustc::middle::privacy::{AccessLevels, AccessLevel};
-use rustc::hir::def::Def;
-use rustc::hir::def_id::{CrateNum, CRATE_DEF_INDEX, DefId};
-use rustc::ty::Visibility;
-use rustc::util::nodemap::FxHashSet;
-
-use std::cell::RefMut;
-
-use crate::clean::{AttributesExt, NestedAttributesExt};
+use rustc_data_structures::fx::FxHashSet;
+use rustc_hir::def::{DefKind, Res};
+use rustc_hir::def_id::{CrateNum, DefId, CRATE_DEF_INDEX};
+use rustc_middle::middle::privacy::{AccessLevel, AccessLevels};
+use rustc_middle::ty::TyCtxt;
 
 // FIXME: this may not be exhaustive, but is sufficient for rustdocs current uses
 
 /// Similar to `librustc_privacy::EmbargoVisitor`, but also takes
 /// specific rustdoc annotations into account (i.e., `doc(hidden)`)
-pub struct LibEmbargoVisitor<'a, 'tcx> {
-    cx: &'a crate::core::DocContext<'tcx>,
+crate struct LibEmbargoVisitor<'a, 'tcx> {
+    tcx: TyCtxt<'tcx>,
     // Accessibility levels for reachable nodes
-    access_levels: RefMut<'a, AccessLevels<DefId>>,
+    access_levels: &'a mut AccessLevels<DefId>,
     // Previous accessibility level, None means unreachable
     prev_level: Option<AccessLevel>,
     // Keeps track of already visited modules, in case a module re-exports its parent
@@ -23,18 +19,16 @@ pub struct LibEmbargoVisitor<'a, 'tcx> {
 }
 
 impl<'a, 'tcx> LibEmbargoVisitor<'a, 'tcx> {
-    pub fn new(
-        cx: &'a crate::core::DocContext<'tcx>
-    ) -> LibEmbargoVisitor<'a, 'tcx> {
+    crate fn new(cx: &'a mut crate::core::DocContext<'tcx>) -> LibEmbargoVisitor<'a, 'tcx> {
         LibEmbargoVisitor {
-            cx,
-            access_levels: RefMut::map(cx.renderinfo.borrow_mut(), |ri| &mut ri.access_levels),
+            tcx: cx.tcx,
+            access_levels: &mut cx.cache.access_levels,
             prev_level: Some(AccessLevel::Public),
-            visited_mods: FxHashSet::default()
+            visited_mods: FxHashSet::default(),
         }
     }
 
-    pub fn visit_lib(&mut self, cnum: CrateNum) {
+    crate fn visit_lib(&mut self, cnum: CrateNum) {
         let did = DefId { krate: cnum, index: CRATE_DEF_INDEX };
         self.update(did, Some(AccessLevel::Public));
         self.visit_mod(did);
@@ -42,7 +36,7 @@ impl<'a, 'tcx> LibEmbargoVisitor<'a, 'tcx> {
 
     // Updates node level and returns the updated level
     fn update(&mut self, did: DefId, level: Option<AccessLevel>) -> Option<AccessLevel> {
-        let is_hidden = self.cx.tcx.get_attrs(did).lists("doc").has_word("hidden");
+        let is_hidden = self.tcx.is_doc_hidden(did);
 
         let old_level = self.access_levels.map.get(&did).cloned();
         // Accessibility levels can only grow
@@ -54,33 +48,30 @@ impl<'a, 'tcx> LibEmbargoVisitor<'a, 'tcx> {
         }
     }
 
-    pub fn visit_mod(&mut self, def_id: DefId) {
+    crate fn visit_mod(&mut self, def_id: DefId) {
         if !self.visited_mods.insert(def_id) {
             return;
         }
 
-        for item in self.cx.tcx.item_children(def_id).iter() {
-            if let Some(def_id) = item.def.opt_def_id() {
-                if self.cx.tcx.def_key(def_id).parent.map_or(false, |d| d == def_id.index) ||
-                    item.vis == Visibility::Public {
-                    self.visit_item(item.def);
+        for item in self.tcx.item_children(def_id).iter() {
+            if let Some(def_id) = item.res.opt_def_id() {
+                if self.tcx.def_key(def_id).parent.map_or(false, |d| d == def_id.index)
+                    || item.vis.is_public()
+                {
+                    self.visit_item(item.res);
                 }
             }
         }
     }
 
-    fn visit_item(&mut self, def: Def) {
-        let def_id = def.def_id();
-        let vis = self.cx.tcx.visibility(def_id);
-        let inherited_item_level = if vis == Visibility::Public {
-            self.prev_level
-        } else {
-            None
-        };
+    fn visit_item(&mut self, res: Res<!>) {
+        let def_id = res.def_id();
+        let vis = self.tcx.visibility(def_id);
+        let inherited_item_level = if vis.is_public() { self.prev_level } else { None };
 
         let item_level = self.update(def_id, inherited_item_level);
 
-        if let Def::Mod(..) = def {
+        if let Res::Def(DefKind::Mod, _) = res {
             let orig_level = self.prev_level;
 
             self.prev_level = item_level;
