@@ -1,7 +1,7 @@
 // compile-flags:-Z unstable-options --show-coverage
-// compile-pass
+// check-pass
 
-#![feature(doc_keyword)]
+#![feature(rustdoc_internals)]
 
 //! the features only used in std also have entries in the table, so make sure those get pulled out
 //! properly as well
