@@ -1,10 +1,10 @@
 // Test that we can't call random fns in a const fn or do other bad things.
 
-#![feature(const_fn_transmute)]
-
 use std::mem::transmute;
 
-fn random() -> u32 { 0 }
+fn random() -> u32 {
+    0
+}
 
 const fn sub(x: &u32) -> usize {
     unsafe { transmute(x) }
@@ -18,12 +18,12 @@ static Y: u32 = 0;
 
 const fn get_Y() -> u32 {
     Y
-        //~^ ERROR E0013
+    //~^ ERROR E0013
 }
 
 const fn get_Y_addr() -> &'static u32 {
     &Y
-        //~^ ERROR E0013
+    //~^ ERROR E0013
 }
 
 const fn get() -> u32 {
