@@ -9,7 +9,7 @@ use std::sync::atomic::{AtomicUsize, Ordering};
 
 use cargo_test_support::install::{cargo_home, exe};
 use cargo_test_support::paths::CargoPathExt;
-use cargo_test_support::registry::Package;
+use cargo_test_support::registry::{self, Package};
 use cargo_test_support::{
     basic_manifest, cargo_process, cross_compile, execs, git, process, project, Execs,
 };
@@ -335,14 +335,16 @@ fn change_features_rebuilds() {
     Package::new("foo", "1.0.0")
         .file(
             "src/main.rs",
-            r#"fn main() {
-            if cfg!(feature = "f1") {
-                println!("f1");
+            r#"
+            fn main() {
+                if cfg!(feature = "f1") {
+                    println!("f1");
+                }
+                if cfg!(feature = "f2") {
+                    println!("f2");
+                }
             }
-            if cfg!(feature = "f2") {
-                println!("f2");
-            }
-        }"#,
+            "#,
         )
         .file(
             "Cargo.toml",
@@ -547,6 +549,7 @@ fn upgrade_git() {
 fn switch_sources() {
     // Installing what appears to be the same thing, but from different
     // sources should reinstall.
+    registry::alt_init();
     pkg("foo", "1.0.0");
     Package::new("foo", "1.0.0")
         .file("src/main.rs", r#"fn main() { println!("alt"); }"#)
@@ -592,20 +595,20 @@ fn multiple_report() {
 [UPDATING] `[..]` index
 [DOWNLOADING] crates ...
 [DOWNLOADED] one v1.0.0 (registry `[..]`)
+[DOWNLOADING] crates ...
+[DOWNLOADED] two v1.0.0 (registry `[..]`)
+[DOWNLOADING] crates ...
+[DOWNLOADED] three v1.0.0 (registry `[..]`)
 [INSTALLING] one v1.0.0
 [COMPILING] one v1.0.0
 [FINISHED] release [optimized] target(s) in [..]
 [INSTALLING] [..]/.cargo/bin/one[EXE]
 [INSTALLED] package `one v1.0.0` (executable `one[EXE]`)
-[DOWNLOADING] crates ...
-[DOWNLOADED] two v1.0.0 (registry `[..]`)
 [INSTALLING] two v1.0.0
 [COMPILING] two v1.0.0
 [FINISHED] release [optimized] target(s) in [..]
 [INSTALLING] [..]/.cargo/bin/two[EXE]
 [INSTALLED] package `two v1.0.0` (executable `two[EXE]`)
-[DOWNLOADING] crates ...
-[DOWNLOADED] three v1.0.0 (registry `[..]`)
 [INSTALLING] three v1.0.0
 [COMPILING] three v1.0.0
 [FINISHED] release [optimized] target(s) in [..]
@@ -691,7 +694,7 @@ fn no_track() {
         .with_stderr(
             "\
 [UPDATING] `[..]` index
-[ERROR] binary `foo[EXE]` already exists in destination
+[ERROR] binary `foo[EXE]` already exists in destination `[..]/.cargo/bin/foo[EXE]`
 Add --force to overwrite
 ",
         )
@@ -800,8 +803,9 @@ fn already_installed_updates_yank_status_on_upgrade() {
     cargo_process("install foo --version=1.0.1")
         .with_status(101)
         .with_stderr_contains(
-            "error: cannot install package `foo`, it has been yanked from registry \
-            `https://github.com/rust-lang/crates.io-index`",
+            "\
+[ERROR] cannot install package `foo`, it has been yanked from registry `crates-io`
+",
         )
         .run();
 
@@ -838,13 +842,13 @@ fn partially_already_installed_does_one_update() {
 [UPDATING] `[..]` index
 [DOWNLOADING] crates ...
 [DOWNLOADED] bar v1.0.0 (registry [..])
+[DOWNLOADING] crates ...
+[DOWNLOADED] baz v1.0.0 (registry [..])
 [INSTALLING] bar v1.0.0
 [COMPILING] bar v1.0.0
 [FINISHED] release [optimized] target(s) in [..]
 [INSTALLING] [CWD]/home/.cargo/bin/bar[EXE]
 [INSTALLED] package `bar v1.0.0` (executable `bar[EXE]`)
-[DOWNLOADING] crates ...
-[DOWNLOADED] baz v1.0.0 (registry [..])
 [INSTALLING] baz v1.0.0
 [COMPILING] baz v1.0.0
 [FINISHED] release [optimized] target(s) in [..]
