@@ -1,16 +1,17 @@
-use cargo;
-use cargo_test_support::project;
+//! Tests for displaying the cargo version.
+
+use cargo_test_support::{cargo_process, project};
 
 #[cargo_test]
 fn simple() {
     let p = project().build();
 
     p.cargo("version")
-        .with_stdout(&format!("{}\n", cargo::version()))
+        .with_stdout(&format!("cargo {}\n", cargo::version()))
         .run();
 
     p.cargo("--version")
-        .with_stdout(&format!("{}\n", cargo::version()))
+        .with_stdout(&format!("cargo {}\n", cargo::version()))
         .run();
 }
 
@@ -33,10 +34,22 @@ fn version_works_with_bad_target_dir() {
         .file(
             ".cargo/config",
             r#"
-            [build]
-            target-dir = 4
-        "#,
+                [build]
+                target-dir = 4
+            "#,
         )
         .build();
     p.cargo("version").run();
 }
+
+#[cargo_test]
+fn verbose() {
+    // This is mainly to check that it doesn't explode.
+    cargo_process("-vV")
+        .with_stdout_contains(&format!("cargo {}", cargo::version()))
+        .with_stdout_contains("host: [..]")
+        .with_stdout_contains("libgit2: [..]")
+        .with_stdout_contains("libcurl: [..]")
+        .with_stdout_contains("os: [..]")
+        .run();
+}
