@@ -1,5 +1,6 @@
 #![feature(rustc_private, bool_to_option, stmt_expr_attributes)]
 
+extern crate rustc_data_structures;
 extern crate rustc_driver;
 extern crate rustc_errors;
 extern crate rustc_hir;
@@ -10,20 +11,21 @@ extern crate rustc_session;
 
 use std::convert::TryFrom;
 use std::env;
+use std::num::NonZeroU64;
 use std::path::PathBuf;
-use std::rc::Rc;
 use std::str::FromStr;
 
 use hex::FromHexError;
 use log::debug;
 
+use rustc_data_structures::sync::Lrc;
 use rustc_driver::Compilation;
 use rustc_errors::emitter::{ColorConfig, HumanReadableErrorType};
-use rustc_hir::def_id::LOCAL_CRATE;
+use rustc_hir::{self as hir, def_id::LOCAL_CRATE, Node};
 use rustc_interface::interface::Config;
 use rustc_middle::{
     middle::exported_symbols::{ExportedSymbol, SymbolExportLevel},
-    ty::{query::Providers, TyCtxt},
+    ty::{query::ExternProviders, TyCtxt},
 };
 use rustc_session::{config::ErrorOutputType, search_paths::PathKind, CtfeBacktrace};
 
@@ -35,13 +37,13 @@ impl rustc_driver::Callbacks for MiriCompilerCalls {
     fn config(&mut self, config: &mut Config) {
         config.override_queries = Some(|_, _, external_providers| {
             external_providers.used_crate_source = |tcx, cnum| {
-                let mut providers = Providers::default();
+                let mut providers = ExternProviders::default();
                 rustc_metadata::provide_extern(&mut providers);
                 let mut crate_source = (providers.used_crate_source)(tcx, cnum);
                 // HACK: rustc will emit "crate ... required to be available in rlib format, but
                 // was not found in this form" errors once we use `tcx.dependency_formats()` if
                 // there's no rlib provided, so setting a dummy path here to workaround those errors.
-                Rc::make_mut(&mut crate_source).rlib = Some((PathBuf::new(), PathKind::All));
+                Lrc::make_mut(&mut crate_source).rlib = Some((PathBuf::new(), PathKind::All));
                 crate_source
             };
         });
@@ -56,8 +58,8 @@ impl rustc_driver::Callbacks for MiriCompilerCalls {
 
         queries.global_ctxt().unwrap().peek_mut().enter(|tcx| {
             init_late_loggers(tcx);
-            let (entry_def_id, _) = if let Some((entry_def, x)) = tcx.entry_fn(()) {
-                (entry_def, x)
+            let (entry_def_id, entry_type) = if let Some(entry_def) = tcx.entry_fn(()) {
+                entry_def
             } else {
                 let output_ty = ErrorOutputType::HumanReadable(HumanReadableErrorType::Default(
                     ColorConfig::Auto,
@@ -77,7 +79,7 @@ impl rustc_driver::Callbacks for MiriCompilerCalls {
                 env::set_current_dir(cwd).unwrap();
             }
 
-            if let Some(return_code) = miri::eval_main(tcx, entry_def_id, config) {
+            if let Some(return_code) = miri::eval_entry(tcx, entry_def_id, entry_type, config) {
                 std::process::exit(
                     i32::try_from(return_code).expect("Return value was too large!"),
                 );
@@ -109,12 +111,27 @@ impl rustc_driver::Callbacks for MiriBeRustCompilerCalls {
                         // https://github.com/rust-lang/rust/blob/2962e7c0089d5c136f4e9600b7abccfbbde4973d/compiler/rustc_codegen_ssa/src/back/symbol_export.rs#L62-L63
                         // https://github.com/rust-lang/rust/blob/2962e7c0089d5c136f4e9600b7abccfbbde4973d/compiler/rustc_codegen_ssa/src/back/symbol_export.rs#L174
                         tcx.reachable_set(()).iter().filter_map(|&local_def_id| {
-                            tcx.codegen_fn_attrs(local_def_id)
-                                .contains_extern_indicator()
-                                .then_some((
-                                    ExportedSymbol::NonGeneric(local_def_id.to_def_id()),
-                                    SymbolExportLevel::C,
-                                ))
+                            // Do the same filtering that rustc does:
+                            // https://github.com/rust-lang/rust/blob/2962e7c0089d5c136f4e9600b7abccfbbde4973d/compiler/rustc_codegen_ssa/src/back/symbol_export.rs#L84-L102
+                            // Otherwise it may cause unexpected behaviours and ICEs
+                            // (https://github.com/rust-lang/rust/issues/86261).
+                            let is_reachable_non_generic = matches!(
+                                tcx.hir().get(tcx.hir().local_def_id_to_hir_id(local_def_id)),
+                                Node::Item(&hir::Item {
+                                    kind: hir::ItemKind::Static(..) | hir::ItemKind::Fn(..),
+                                    ..
+                                }) | Node::ImplItem(&hir::ImplItem {
+                                    kind: hir::ImplItemKind::Fn(..),
+                                    ..
+                                })
+                                if !tcx.generics_of(local_def_id).requires_monomorphization(tcx)
+                            );
+                            (is_reachable_non_generic
+                                && tcx.codegen_fn_attrs(local_def_id).contains_extern_indicator())
+                            .then_some((
+                                ExportedSymbol::NonGeneric(local_def_id.to_def_id()),
+                                SymbolExportLevel::C,
+                            ))
                         }),
                     )
                 }
@@ -147,11 +164,14 @@ fn init_late_loggers(tcx: TyCtxt<'_>) {
             // used for everything, we only apply it to the parts of rustc that are
             // CTFE-related. Otherwise, we use it verbatim for `RUSTC_LOG`.
             // This way, if you set `MIRI_LOG=trace`, you get only the right parts of
-            // rustc traced, but you can also do `MIRI_LOG=miri=trace,rustc_mir::interpret=debug`.
+            // rustc traced, but you can also do `MIRI_LOG=miri=trace,rustc_const_eval::interpret=debug`.
             if log::Level::from_str(&var).is_ok() {
                 env::set_var(
                     "RUSTC_LOG",
-                    &format!("rustc_middle::mir::interpret={0},rustc_mir::interpret={0}", var),
+                    &format!(
+                        "rustc_middle::mir::interpret={0},rustc_const_eval::interpret={0}",
+                        var
+                    ),
                 );
             } else {
                 env::set_var("RUSTC_LOG", &var);
@@ -188,9 +208,12 @@ fn compile_time_sysroot() -> Option<String> {
     let toolchain = option_env!("RUSTUP_TOOLCHAIN").or(option_env!("MULTIRUST_TOOLCHAIN"));
     Some(match (home, toolchain) {
         (Some(home), Some(toolchain)) => format!("{}/toolchains/{}", home, toolchain),
-        _ => option_env!("RUST_SYSROOT")
-            .expect("To build Miri without rustup, set the `RUST_SYSROOT` env var at build time")
-            .to_owned(),
+        _ =>
+            option_env!("RUST_SYSROOT")
+                .expect(
+                    "To build Miri without rustup, set the `RUST_SYSROOT` env var at build time",
+                )
+                .to_owned(),
     })
 }
 
@@ -290,6 +313,9 @@ fn main() {
                 "-Zmiri-symbolic-alignment-check" => {
                     miri_config.check_alignment = miri::AlignmentCheck::Symbolic;
                 }
+                "-Zmiri-check-number-validity" => {
+                    miri_config.check_number_validity = true;
+                }
                 "-Zmiri-disable-abi-check" => {
                     miri_config.check_abi = false;
                 }
@@ -321,9 +347,10 @@ fn main() {
                         "warn" => miri::IsolatedOp::Reject(miri::RejectOpWith::Warning),
                         "warn-nobacktrace" =>
                             miri::IsolatedOp::Reject(miri::RejectOpWith::WarningWithoutBacktrace),
-                        _ => panic!(
-                            "-Zmiri-isolation-error must be `abort`, `hide`, `warn`, or `warn-nobacktrace`"
-                        ),
+                        _ =>
+                            panic!(
+                                "-Zmiri-isolation-error must be `abort`, `hide`, `warn`, or `warn-nobacktrace`"
+                            ),
                     };
                 }
                 "-Zmiri-ignore-leaks" => {
@@ -332,8 +359,14 @@ fn main() {
                 "-Zmiri-panic-on-unsupported" => {
                     miri_config.panic_on_unsupported = true;
                 }
+                "-Zmiri-tag-raw-pointers" => {
+                    miri_config.tag_raw = true;
+                }
                 "-Zmiri-track-raw-pointers" => {
-                    miri_config.track_raw = true;
+                    eprintln!(
+                        "WARNING: -Zmiri-track-raw-pointers has been renamed to -Zmiri-tag-raw-pointers, the old name is deprecated."
+                    );
+                    miri_config.tag_raw = true;
                 }
                 "--" => {
                     after_dashdash = true;
@@ -368,10 +401,11 @@ fn main() {
                     let id: u64 =
                         match arg.strip_prefix("-Zmiri-track-pointer-tag=").unwrap().parse() {
                             Ok(id) => id,
-                            Err(err) => panic!(
-                                "-Zmiri-track-pointer-tag requires a valid `u64` argument: {}",
-                                err
-                            ),
+                            Err(err) =>
+                                panic!(
+                                    "-Zmiri-track-pointer-tag requires a valid `u64` argument: {}",
+                                    err
+                                ),
                         };
                     if let Some(id) = miri::PtrId::new(id) {
                         miri_config.tracked_pointer_tag = Some(id);
@@ -392,11 +426,16 @@ fn main() {
                     }
                 }
                 arg if arg.starts_with("-Zmiri-track-alloc-id=") => {
-                    let id: u64 = match arg.strip_prefix("-Zmiri-track-alloc-id=").unwrap().parse()
+                    let id = match arg
+                        .strip_prefix("-Zmiri-track-alloc-id=")
+                        .unwrap()
+                        .parse()
+                        .ok()
+                        .and_then(NonZeroU64::new)
                     {
-                        Ok(id) => id,
-                        Err(err) =>
-                            panic!("-Zmiri-track-alloc-id requires a valid `u64` argument: {}", err),
+                        Some(id) => id,
+                        None =>
+                            panic!("-Zmiri-track-alloc-id requires a valid non-zero `u64` argument"),
                     };
                     miri_config.tracked_alloc_id = Some(miri::AllocId(id));
                 }
@@ -407,13 +446,15 @@ fn main() {
                         .parse::<f64>()
                     {
                         Ok(rate) if rate >= 0.0 && rate <= 1.0 => rate,
-                        Ok(_) => panic!(
-                            "-Zmiri-compare-exchange-weak-failure-rate must be between `0.0` and `1.0`"
-                        ),
-                        Err(err) => panic!(
-                            "-Zmiri-compare-exchange-weak-failure-rate requires a `f64` between `0.0` and `1.0`: {}",
-                            err
-                        ),
+                        Ok(_) =>
+                            panic!(
+                                "-Zmiri-compare-exchange-weak-failure-rate must be between `0.0` and `1.0`"
+                            ),
+                        Err(err) =>
+                            panic!(
+                                "-Zmiri-compare-exchange-weak-failure-rate requires a `f64` between `0.0` and `1.0`: {}",
+                                err
+                            ),
                     };
                     miri_config.cmpxchg_weak_failure_rate = rate;
                 }
