@@ -7,6 +7,7 @@ use log::trace;
 use rustc_middle::ty::{self, TyCtxt};
 use rustc_span::{source_map::DUMMY_SP, Span, SpanData, Symbol};
 
+use crate::stacked_borrows::{AccessKind, SbTag};
 use crate::*;
 
 /// Details of premature program termination.
@@ -26,6 +27,10 @@ pub enum TerminationInfo {
         second: SpanData,
         second_crate: Symbol,
     },
+    SymbolShimClashing {
+        link_name: Symbol,
+        span: SpanData,
+    },
 }
 
 impl fmt::Display for TerminationInfo {
@@ -39,6 +44,12 @@ impl fmt::Display for TerminationInfo {
             Deadlock => write!(f, "the evaluated program deadlocked"),
             MultipleSymbolDefinitions { link_name, .. } =>
                 write!(f, "multiple definitions of symbol `{}`", link_name),
+            SymbolShimClashing { link_name, .. } =>
+                write!(
+                    f,
+                    "found `{}` symbol definition that clashes with a built-in shim",
+                    link_name
+                ),
         }
     }
 }
@@ -48,7 +59,9 @@ impl MachineStopType for TerminationInfo {}
 /// Miri specific diagnostics
 pub enum NonHaltingDiagnostic {
     CreatedPointerTag(NonZeroU64),
-    PoppedPointerTag(Item),
+    /// This `Item` was popped from the borrow stack, either due to a grant of
+    /// `AccessKind` to `SbTag` or a deallocation when the second argument is `None`.
+    PoppedPointerTag(Item, Option<(SbTag, AccessKind)>),
     CreatedCallId(CallId),
     CreatedAlloc(AllocId),
     FreedAlloc(AllocId),
@@ -79,7 +92,7 @@ pub fn report_error<'tcx, 'mir>(
                 UnsupportedInIsolation(_) => Some("unsupported operation"),
                 ExperimentalUb { .. } => Some("Undefined Behavior"),
                 Deadlock => Some("deadlock"),
-                MultipleSymbolDefinitions { .. } => None,
+                MultipleSymbolDefinitions { .. } | SymbolShimClashing { .. } => None,
             };
             #[rustfmt::skip]
             let helps = match info {
@@ -98,6 +111,8 @@ pub fn report_error<'tcx, 'mir>(
                         (Some(*first), format!("it's first defined here, in crate `{}`", first_crate)),
                         (Some(*second), format!("then it's defined here again, in crate `{}`", second_crate)),
                     ],
+                SymbolShimClashing { link_name, span } =>
+                    vec![(Some(*span), format!("the `{}` symbol is defined here", link_name))],
                 _ => vec![],
             };
             (title, helps)
@@ -120,9 +135,7 @@ pub fn report_error<'tcx, 'mir>(
             };
             #[rustfmt::skip]
             let helps = match e.kind() {
-                Unsupported(UnsupportedOpInfo::NoMirFor(..)) =>
-                    vec![(None, format!("make sure to use a Miri sysroot, which you can prepare with `cargo miri setup`"))],
-                Unsupported(UnsupportedOpInfo::ReadBytesAsPointer | UnsupportedOpInfo::ThreadLocalStatic(_) | UnsupportedOpInfo::ReadExternStatic(_)) =>
+                Unsupported(UnsupportedOpInfo::ThreadLocalStatic(_) | UnsupportedOpInfo::ReadExternStatic(_)) =>
                     panic!("Error should never be raised by Miri: {:?}", e.kind()),
                 Unsupported(_) =>
                     vec![(None, format!("this is likely not a bug in the program; it indicates that the program performed an operation that the interpreter does not support"))],
@@ -311,12 +324,25 @@ pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx
                 use NonHaltingDiagnostic::*;
                 let msg = match e {
                     CreatedPointerTag(tag) => format!("created tag {:?}", tag),
-                    PoppedPointerTag(item) => format!("popped tracked tag for item {:?}", item),
+                    PoppedPointerTag(item, tag) =>
+                        match tag {
+                            None =>
+                                format!(
+                                    "popped tracked tag for item {:?} due to deallocation",
+                                    item
+                                ),
+                            Some((tag, access)) => {
+                                format!(
+                                    "popped tracked tag for item {:?} due to {:?} access for {:?}",
+                                    item, access, tag
+                                )
+                            }
+                        },
                     CreatedCallId(id) => format!("function call with id {}", id),
                     CreatedAlloc(AllocId(id)) => format!("created allocation with id {}", id),
                     FreedAlloc(AllocId(id)) => format!("freed allocation with id {}", id),
                     RejectedIsolatedOp(ref op) =>
-                        format!("`{}` was made to return an error due to isolation", op),
+                        format!("{} was made to return an error due to isolation", op),
                 };
 
                 let (title, diag_level) = match e {
