@@ -5,10 +5,15 @@ use std::time::Duration;
 
 use log::trace;
 
-use rustc_middle::mir;
-use rustc_middle::ty::{self, List, TyCtxt, layout::TyAndLayout};
 use rustc_hir::def_id::{DefId, CRATE_DEF_INDEX};
-use rustc_target::abi::{LayoutOf, Size, FieldsShape, Variants};
+use rustc_middle::mir;
+use rustc_middle::ty::{
+    self,
+    layout::{LayoutOf, TyAndLayout},
+    List, TyCtxt,
+};
+use rustc_span::Symbol;
+use rustc_target::abi::{Align, FieldsShape, Size, Variants};
 use rustc_target::spec::abi::Abi;
 
 use rand::RngCore;
@@ -19,10 +24,8 @@ impl<'mir, 'tcx: 'mir> EvalContextExt<'mir, 'tcx> for crate::MiriEvalContext<'mi
 
 /// Gets an instance for a path.
 fn try_resolve_did<'mir, 'tcx>(tcx: TyCtxt<'tcx>, path: &[&str]) -> Option<DefId> {
-    tcx.crates()
-        .iter()
-        .find(|&&krate| tcx.original_crate_name(krate).as_str() == path[0])
-        .and_then(|krate| {
+    tcx.crates(()).iter().find(|&&krate| tcx.crate_name(krate).as_str() == path[0]).and_then(
+        |krate| {
             let krate = DefId { krate: *krate, index: CRATE_DEF_INDEX };
             let mut items = tcx.item_children(krate);
             let mut path_it = path.iter().skip(1).peekable();
@@ -40,7 +43,8 @@ fn try_resolve_did<'mir, 'tcx>(tcx: TyCtxt<'tcx>, path: &[&str]) -> Option<DefId
                 }
             }
             None
-        })
+        },
+    )
 }
 
 pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx> {
@@ -53,23 +57,18 @@ pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx
 
     /// Evaluates the scalar at the specified path. Returns Some(val)
     /// if the path could be resolved, and None otherwise
-    fn eval_path_scalar(
-        &mut self,
-        path: &[&str],
-    ) -> InterpResult<'tcx, ScalarMaybeUninit<Tag>> {
+    fn eval_path_scalar(&mut self, path: &[&str]) -> InterpResult<'tcx, Scalar<Tag>> {
         let this = self.eval_context_mut();
         let instance = this.resolve_path(path);
         let cid = GlobalId { instance, promoted: None };
         let const_val = this.eval_to_allocation(cid)?;
         let const_val = this.read_scalar(&const_val.into())?;
-        return Ok(const_val);
+        return Ok(const_val.check_init()?);
     }
 
     /// Helper function to get a `libc` constant as a `Scalar`.
     fn eval_libc(&mut self, name: &str) -> InterpResult<'tcx, Scalar<Tag>> {
-        self.eval_context_mut()
-            .eval_path_scalar(&["libc", name])?
-            .check_init()
+        self.eval_context_mut().eval_path_scalar(&["libc", name])
     }
 
     /// Helper function to get a `libc` constant as an `i32`.
@@ -80,12 +79,10 @@ pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx
 
     /// Helper function to get a `windows` constant as a `Scalar`.
     fn eval_windows(&mut self, module: &str, name: &str) -> InterpResult<'tcx, Scalar<Tag>> {
-        self.eval_context_mut()
-            .eval_path_scalar(&["std", "sys", "windows", module, name])?
-            .check_init()
+        self.eval_context_mut().eval_path_scalar(&["std", "sys", "windows", module, name])
     }
 
-    /// Helper function to get a `windows` constant as an `u64`.
+    /// Helper function to get a `windows` constant as a `u64`.
     fn eval_windows_u64(&mut self, module: &str, name: &str) -> InterpResult<'tcx, u64> {
         // TODO: Cache the result.
         self.eval_windows(module, name)?.to_u64()
@@ -101,7 +98,9 @@ pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx
     /// Helper function to get the `TyAndLayout` of a `windows` type
     fn windows_ty_layout(&mut self, name: &str) -> InterpResult<'tcx, TyAndLayout<'tcx>> {
         let this = self.eval_context_mut();
-        let ty = this.resolve_path(&["std", "sys", "windows", "c", name]).ty(*this.tcx, ty::ParamEnv::reveal_all());
+        let ty = this
+            .resolve_path(&["std", "sys", "windows", "c", name])
+            .ty(*this.tcx, ty::ParamEnv::reveal_all());
         this.layout_of(ty)
     }
 
@@ -110,17 +109,11 @@ pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx
         self.eval_context_mut().write_scalar(Scalar::from_int(0, dest.layout.size), dest)
     }
 
-    /// Test if this immediate equals 0.
-    fn is_null(&self, val: Scalar<Tag>) -> InterpResult<'tcx, bool> {
+    /// Test if this pointer equals 0.
+    fn ptr_is_null(&self, ptr: Pointer<Option<Tag>>) -> InterpResult<'tcx, bool> {
         let this = self.eval_context_ref();
         let null = Scalar::null_ptr(this);
-        this.ptr_eq(val, null)
-    }
-
-    /// Turn a Scalar into an Option<NonNullScalar>
-    fn test_null(&self, val: Scalar<Tag>) -> InterpResult<'tcx, Option<Scalar<Tag>>> {
-        let this = self.eval_context_ref();
-        Ok(if this.is_null(val)? { None } else { Some(val) })
+        this.ptr_eq(Scalar::from_maybe_pointer(ptr, this), null)
     }
 
     /// Get the `Place` for a local
@@ -131,7 +124,7 @@ pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx
     }
 
     /// Generate some random bytes, and write them to `dest`.
-    fn gen_random(&mut self, ptr: Scalar<Tag>, len: u64) -> InterpResult<'tcx> {
+    fn gen_random(&mut self, ptr: Pointer<Option<Tag>>, len: u64) -> InterpResult<'tcx> {
         // Some programs pass in a null pointer and a length of 0
         // to their platform's random-generation function (e.g. getrandom())
         // on Linux. For compatibility with these programs, we don't perform
@@ -144,7 +137,7 @@ pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx
 
         let mut data = vec![0; usize::try_from(len).unwrap()];
 
-        if this.machine.communicate {
+        if this.machine.communicate() {
             // Fill the buffer using the host's rng.
             getrandom::getrandom(&mut data)
                 .map_err(|err| err_unsup_format!("host getrandom failed: {}", err))?;
@@ -169,8 +162,12 @@ pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx
         let this = self.eval_context_mut();
         let param_env = ty::ParamEnv::reveal_all(); // in Miri this is always the param_env we use... and this.param_env is private.
         let callee_abi = f.ty(*this.tcx, param_env).fn_sig(*this.tcx).abi();
-        if callee_abi != caller_abi {
-            throw_ub_format!("calling a function with ABI {} using caller ABI {}", callee_abi.name(), caller_abi.name())
+        if this.machine.enforce_abi && callee_abi != caller_abi {
+            throw_ub_format!(
+                "calling a function with ABI {} using caller ABI {}",
+                callee_abi.name(),
+                caller_abi.name()
+            )
         }
 
         // Push frame.
@@ -181,9 +178,9 @@ pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx
         let mut callee_args = this.frame().body.args_iter();
         for arg in args {
             let callee_arg = this.local_place(
-                callee_args.next().ok_or_else(||
-                    err_ub_format!("callee has fewer arguments than expected")
-                )?
+                callee_args
+                    .next()
+                    .ok_or_else(|| err_ub_format!("callee has fewer arguments than expected"))?,
             )?;
             this.write_immediate(*arg, &callee_arg)?;
         }
@@ -194,13 +191,16 @@ pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx
         Ok(())
     }
 
-    /// Visits the memory covered by `place`, sensitive to freezing: the 3rd parameter
-    /// will be true if this is frozen, false if this is in an `UnsafeCell`.
+    /// Visits the memory covered by `place`, sensitive to freezing: the 2nd parameter
+    /// of `action` will be true if this is frozen, false if this is in an `UnsafeCell`.
+    /// The range is relative to `place`.
+    ///
+    /// Assumes that the `place` has a proper pointer in it.
     fn visit_freeze_sensitive(
         &self,
         place: &MPlaceTy<'tcx, Tag>,
         size: Size,
-        mut action: impl FnMut(Pointer<Tag>, Size, bool) -> InterpResult<'tcx>,
+        mut action: impl FnMut(AllocRange, bool) -> InterpResult<'tcx>,
     ) -> InterpResult<'tcx> {
         let this = self.eval_context_ref();
         trace!("visit_frozen(place={:?}, size={:?})", *place, size);
@@ -213,29 +213,33 @@ pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx
         // Store how far we proceeded into the place so far. Everything to the left of
         // this offset has already been handled, in the sense that the frozen parts
         // have had `action` called on them.
-        let mut end_ptr = place.ptr.assert_ptr();
+        let ptr = place.ptr.into_pointer_or_addr().unwrap();
+        let start_offset = ptr.into_parts().1 as Size; // we just compare offsets, the abs. value never matters
+        let mut cur_offset = start_offset;
         // Called when we detected an `UnsafeCell` at the given offset and size.
-        // Calls `action` and advances `end_ptr`.
-        let mut unsafe_cell_action = |unsafe_cell_ptr: Scalar<Tag>, unsafe_cell_size: Size| {
-            let unsafe_cell_ptr = unsafe_cell_ptr.assert_ptr();
-            debug_assert_eq!(unsafe_cell_ptr.alloc_id, end_ptr.alloc_id);
-            debug_assert_eq!(unsafe_cell_ptr.tag, end_ptr.tag);
+        // Calls `action` and advances `cur_ptr`.
+        let mut unsafe_cell_action = |unsafe_cell_ptr: Pointer<Option<Tag>>,
+                                      unsafe_cell_size: Size| {
+            let unsafe_cell_ptr = unsafe_cell_ptr.into_pointer_or_addr().unwrap();
+            debug_assert_eq!(unsafe_cell_ptr.provenance, ptr.provenance);
             // We assume that we are given the fields in increasing offset order,
             // and nothing else changes.
-            let unsafe_cell_offset = unsafe_cell_ptr.offset;
-            let end_offset = end_ptr.offset;
-            assert!(unsafe_cell_offset >= end_offset);
-            let frozen_size = unsafe_cell_offset - end_offset;
-            // Everything between the end_ptr and this `UnsafeCell` is frozen.
+            let unsafe_cell_offset = unsafe_cell_ptr.into_parts().1 as Size; // we just compare offsets, the abs. value never matters
+            assert!(unsafe_cell_offset >= cur_offset);
+            let frozen_size = unsafe_cell_offset - cur_offset;
+            // Everything between the cur_ptr and this `UnsafeCell` is frozen.
             if frozen_size != Size::ZERO {
-                action(end_ptr, frozen_size, /*frozen*/ true)?;
+                action(alloc_range(cur_offset - start_offset, frozen_size), /*frozen*/ true)?;
             }
+            cur_offset += frozen_size;
             // This `UnsafeCell` is NOT frozen.
             if unsafe_cell_size != Size::ZERO {
-                action(unsafe_cell_ptr, unsafe_cell_size, /*frozen*/ false)?;
+                action(
+                    alloc_range(cur_offset - start_offset, unsafe_cell_size),
+                    /*frozen*/ false,
+                )?;
             }
-            // Update end end_ptr.
-            end_ptr = unsafe_cell_ptr.wrapping_offset(unsafe_cell_size, this);
+            cur_offset += unsafe_cell_size;
             // Done
             Ok(())
         };
@@ -263,7 +267,7 @@ pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx
         }
         // The part between the end_ptr and the end of the place is also frozen.
         // So pretend there is a 0-sized `UnsafeCell` at the end.
-        unsafe_cell_action(place.ptr.ptr_wrapping_offset(size, this), Size::ZERO)?;
+        unsafe_cell_action(place.ptr.wrapping_offset(size, this), Size::ZERO)?;
         // Done!
         return Ok(());
 
@@ -346,7 +350,10 @@ pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx
                         // Gather the subplaces and sort them before visiting.
                         let mut places =
                             fields.collect::<InterpResult<'tcx, Vec<MPlaceTy<'tcx, Tag>>>>()?;
-                        places.sort_by_key(|place| place.ptr.assert_ptr().offset);
+                        // we just compare offsets, the abs. value never matters
+                        places.sort_by_key(|place| {
+                            place.ptr.into_pointer_or_addr().unwrap().into_parts().1 as Size
+                        });
                         self.walk_aggregate(place, places.into_iter().map(Ok))
                     }
                     FieldsShape::Union { .. } | FieldsShape::Primitive => {
@@ -356,7 +363,11 @@ pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx
                 }
             }
 
-            fn visit_union(&mut self, _v: &MPlaceTy<'tcx, Tag>, _fields: NonZeroUsize) -> InterpResult<'tcx> {
+            fn visit_union(
+                &mut self,
+                _v: &MPlaceTy<'tcx, Tag>,
+                _fields: NonZeroUsize,
+            ) -> InterpResult<'tcx> {
                 bug!("we should have already handled unions in `visit_value`")
             }
         }
@@ -374,9 +385,9 @@ pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx
         let mut offset = Size::from_bytes(0);
 
         for &imm in imms {
-            this.write_immediate_to_mplace(
+            this.write_immediate(
                 *imm,
-                &place.offset(offset, MemPlaceMeta::None, imm.layout, &*this.tcx)?,
+                &place.offset(offset, MemPlaceMeta::None, imm.layout, &*this.tcx)?.into(),
             )?;
             offset += imm.layout.size;
         }
@@ -387,12 +398,32 @@ pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx
     /// disabled. It returns an error using the `name` of the foreign function if this is not the
     /// case.
     fn check_no_isolation(&self, name: &str) -> InterpResult<'tcx> {
-        if !self.eval_context_ref().machine.communicate {
-            isolation_error(name)?;
+        if !self.eval_context_ref().machine.communicate() {
+            self.reject_in_isolation(name, RejectOpWith::Abort)?;
         }
         Ok(())
     }
 
+    /// Helper function used inside the shims of foreign functions which reject the op
+    /// when isolation is enabled. It is used to print a warning/backtrace about the rejection.
+    fn reject_in_isolation(&self, op_name: &str, reject_with: RejectOpWith) -> InterpResult<'tcx> {
+        let this = self.eval_context_ref();
+        match reject_with {
+            RejectOpWith::Abort => isolation_abort_error(op_name),
+            RejectOpWith::WarningWithoutBacktrace => {
+                this.tcx
+                    .sess
+                    .warn(&format!("{} was made to return an error due to isolation", op_name));
+                Ok(())
+            }
+            RejectOpWith::Warning => {
+                register_diagnostic(NonHaltingDiagnostic::RejectedIsolatedOp(op_name.to_string()));
+                Ok(())
+            }
+            RejectOpWith::NoWarning => Ok(()), // no warning
+        }
+    }
+
     /// Helper function used inside the shims of foreign functions to assert that the target OS
     /// is `target_os`. It panics showing a message with the `name` of the foreign function
     /// if this is not the case.
@@ -415,7 +446,7 @@ pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx
         } else {
             // Allocate new place, set initial value to 0.
             let errno_layout = this.machine.layouts.u32;
-            let errno_place = this.allocate(errno_layout, MiriMemoryKind::Machine.into());
+            let errno_place = this.allocate(errno_layout, MiriMemoryKind::Machine.into())?;
             this.write_scalar(Scalar::from_u32(0), &errno_place.into())?;
             this.active_thread_mut().last_error = Some(errno_place);
             Ok(errno_place)
@@ -436,15 +467,15 @@ pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx
         this.read_scalar(&errno_place.into())?.check_init()
     }
 
-    /// Sets the last OS error using a `std::io::Error`. This function tries to produce the most
+    /// Sets the last OS error using a `std::io::ErrorKind`. This function tries to produce the most
     /// similar OS error from the `std::io::ErrorKind` and sets it as the last OS error.
-    fn set_last_error_from_io_error(&mut self, e: std::io::Error) -> InterpResult<'tcx> {
+    fn set_last_error_from_io_error(&mut self, err_kind: std::io::ErrorKind) -> InterpResult<'tcx> {
         use std::io::ErrorKind::*;
         let this = self.eval_context_mut();
         let target = &this.tcx.sess.target;
         let target_os = &target.os;
-        let last_error = if target.os_family == Some("unix".to_owned()) {
-            this.eval_libc(match e.kind() {
+        let last_error = if target.families.contains(&"unix".to_owned()) {
+            this.eval_libc(match err_kind {
                 ConnectionRefused => "ECONNREFUSED",
                 ConnectionReset => "ECONNRESET",
                 PermissionDenied => "EPERM",
@@ -460,17 +491,31 @@ pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx
                 AlreadyExists => "EEXIST",
                 WouldBlock => "EWOULDBLOCK",
                 _ => {
-                    throw_unsup_format!("io error {} cannot be transformed into a raw os error", e)
+                    throw_unsup_format!(
+                        "io error {:?} cannot be translated into a raw os error",
+                        err_kind
+                    )
                 }
             })?
-        } else if target_os == "windows" {
+        } else if target.families.contains(&"windows".to_owned()) {
             // FIXME: we have to finish implementing the Windows equivalent of this.
-            this.eval_windows("c", match e.kind() {
-                NotFound => "ERROR_FILE_NOT_FOUND",
-                _ => throw_unsup_format!("io error {} cannot be transformed into a raw os error", e)
-            })?
+            this.eval_windows(
+                "c",
+                match err_kind {
+                    NotFound => "ERROR_FILE_NOT_FOUND",
+                    PermissionDenied => "ERROR_ACCESS_DENIED",
+                    _ =>
+                        throw_unsup_format!(
+                            "io error {:?} cannot be translated into a raw os error",
+                            err_kind
+                        ),
+                },
+            )?
         } else {
-            throw_unsup_format!("setting the last OS error from an io::Error is unsupported for {}.", target_os)
+            throw_unsup_format!(
+                "setting the last OS error from an io::Error is unsupported for {}.",
+                target_os
+            )
         };
         this.set_last_error(last_error)
     }
@@ -488,7 +533,7 @@ pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx
         match result {
             Ok(ok) => Ok(ok),
             Err(e) => {
-                self.eval_context_mut().set_last_error_from_io_error(e)?;
+                self.eval_context_mut().set_last_error_from_io_error(e.kind())?;
                 Ok((-1).into())
             }
         }
@@ -528,12 +573,8 @@ pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx
     /// Parse a `timespec` struct and return it as a `std::time::Duration`. It returns `None`
     /// if the value in the `timespec` struct is invalid. Some libc functions will return
     /// `EINVAL` in this case.
-    fn read_timespec(
-        &mut self,
-        timespec_ptr_op: &OpTy<'tcx, Tag>,
-    ) -> InterpResult<'tcx, Option<Duration>> {
+    fn read_timespec(&mut self, tp: &MPlaceTy<'tcx, Tag>) -> InterpResult<'tcx, Option<Duration>> {
         let this = self.eval_context_mut();
-        let tp = this.deref_operand(timespec_ptr_op)?;
         let seconds_place = this.mplace_field(&tp, 0)?;
         let seconds_scalar = this.read_scalar(&seconds_place.into())?;
         let seconds = seconds_scalar.to_machine_isize(this)?;
@@ -553,27 +594,145 @@ pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx
             Duration::new(seconds, nanoseconds)
         })
     }
+
+    fn read_c_str<'a>(&'a self, ptr: Pointer<Option<Tag>>) -> InterpResult<'tcx, &'a [u8]>
+    where
+        'tcx: 'a,
+        'mir: 'a,
+    {
+        let this = self.eval_context_ref();
+        let size1 = Size::from_bytes(1);
+
+        // Step 1: determine the length.
+        let mut len = Size::ZERO;
+        loop {
+            // FIXME: We are re-getting the allocation each time around the loop.
+            // Would be nice if we could somehow "extend" an existing AllocRange.
+            let alloc = this.memory.get(ptr.offset(len, this)?.into(), size1, Align::ONE)?.unwrap(); // not a ZST, so we will get a result
+            let byte = alloc.read_scalar(alloc_range(Size::ZERO, size1))?.to_u8()?;
+            if byte == 0 {
+                break;
+            } else {
+                len = len + size1;
+            }
+        }
+
+        // Step 2: get the bytes.
+        this.memory.read_bytes(ptr.into(), len)
+    }
+
+    fn read_wide_str(&self, mut ptr: Pointer<Option<Tag>>) -> InterpResult<'tcx, Vec<u16>> {
+        let this = self.eval_context_ref();
+        let size2 = Size::from_bytes(2);
+        let align2 = Align::from_bytes(2).unwrap();
+
+        let mut wchars = Vec::new();
+        loop {
+            // FIXME: We are re-getting the allocation each time around the loop.
+            // Would be nice if we could somehow "extend" an existing AllocRange.
+            let alloc = this.memory.get(ptr.into(), size2, align2)?.unwrap(); // not a ZST, so we will get a result
+            let wchar = alloc.read_scalar(alloc_range(Size::ZERO, size2))?.to_u16()?;
+            if wchar == 0 {
+                break;
+            } else {
+                wchars.push(wchar);
+                ptr = ptr.offset(size2, this)?;
+            }
+        }
+
+        Ok(wchars)
+    }
+
+    /// Check that the ABI is what we expect.
+    fn check_abi<'a>(&self, abi: Abi, exp_abi: Abi) -> InterpResult<'a, ()> {
+        if self.eval_context_ref().machine.enforce_abi && abi != exp_abi {
+            throw_ub_format!(
+                "calling a function with ABI {} using caller ABI {}",
+                exp_abi.name(),
+                abi.name()
+            )
+        }
+        Ok(())
+    }
+
+    fn frame_in_std(&self) -> bool {
+        let this = self.eval_context_ref();
+        this.tcx.lang_items().start_fn().map_or(false, |start_fn| {
+            this.tcx.def_path(this.frame().instance.def_id()).krate
+                == this.tcx.def_path(start_fn).krate
+        })
+    }
+
+    /// Handler that should be called when unsupported functionality is encountered.
+    /// This function will either panic within the context of the emulated application
+    /// or return an error in the Miri process context
+    ///
+    /// Return value of `Ok(bool)` indicates whether execution should continue.
+    fn handle_unsupported<S: AsRef<str>>(&mut self, error_msg: S) -> InterpResult<'tcx, ()> {
+        let this = self.eval_context_mut();
+        if this.machine.panic_on_unsupported {
+            // message is slightly different here to make automated analysis easier
+            let error_msg = format!("unsupported Miri functionality: {}", error_msg.as_ref());
+            this.start_panic(error_msg.as_ref(), StackPopUnwind::Skip)?;
+            return Ok(());
+        } else {
+            throw_unsup_format!("{}", error_msg.as_ref());
+        }
+    }
+
+    fn check_abi_and_shim_symbol_clash(
+        &mut self,
+        abi: Abi,
+        exp_abi: Abi,
+        link_name: Symbol,
+    ) -> InterpResult<'tcx, ()> {
+        self.check_abi(abi, exp_abi)?;
+        if let Some((body, _)) = self.eval_context_mut().lookup_exported_symbol(link_name)? {
+            throw_machine_stop!(TerminationInfo::SymbolShimClashing {
+                link_name,
+                span: body.span.data(),
+            })
+        }
+        Ok(())
+    }
+
+    fn check_shim<'a, const N: usize>(
+        &mut self,
+        abi: Abi,
+        exp_abi: Abi,
+        link_name: Symbol,
+        args: &'a [OpTy<'tcx, Tag>],
+    ) -> InterpResult<'tcx, &'a [OpTy<'tcx, Tag>; N]>
+    where
+        &'a [OpTy<'tcx, Tag>; N]: TryFrom<&'a [OpTy<'tcx, Tag>]>,
+    {
+        self.check_abi_and_shim_symbol_clash(abi, exp_abi, link_name)?;
+        check_arg_count(args)
+    }
+
+    /// Mark a machine allocation that was just created as immutable.
+    fn mark_immutable(&mut self, mplace: &MemPlace<Tag>) {
+        let this = self.eval_context_mut();
+        this.memory
+            .mark_immutable(mplace.ptr.into_pointer_or_addr().unwrap().provenance.alloc_id)
+            .unwrap();
+    }
 }
 
 /// Check that the number of args is what we expect.
-pub fn check_arg_count<'a, 'tcx, const N: usize>(args: &'a [OpTy<'tcx, Tag>]) -> InterpResult<'tcx, &'a [OpTy<'tcx, Tag>; N]>
-    where &'a [OpTy<'tcx, Tag>; N]: TryFrom<&'a [OpTy<'tcx, Tag>]> {
+pub fn check_arg_count<'a, 'tcx, const N: usize>(
+    args: &'a [OpTy<'tcx, Tag>],
+) -> InterpResult<'tcx, &'a [OpTy<'tcx, Tag>; N]>
+where
+    &'a [OpTy<'tcx, Tag>; N]: TryFrom<&'a [OpTy<'tcx, Tag>]>,
+{
     if let Ok(ops) = args.try_into() {
         return Ok(ops);
     }
     throw_ub_format!("incorrect number of arguments: got {}, expected {}", args.len(), N)
 }
 
-/// Check that the ABI is what we expect.
-pub fn check_abi<'a>(abi: Abi, exp_abi: Abi) -> InterpResult<'a, ()> {
-    if abi == exp_abi {
-        Ok(())
-    } else {
-        throw_ub_format!("calling a function with ABI {} using caller ABI {}", exp_abi.name(), abi.name())
-    }
-}
-
-pub fn isolation_error(name: &str) -> InterpResult<'static> {
+pub fn isolation_abort_error(name: &str) -> InterpResult<'static> {
     throw_machine_stop!(TerminationInfo::UnsupportedInIsolation(format!(
         "{} not available when isolation is enabled",
         name,
