@@ -7,7 +7,6 @@ use std::fmt;
 use std::num::NonZeroU64;
 use std::time::Instant;
 
-use log::trace;
 use rand::rngs::StdRng;
 use rand::SeedableRng;
 
@@ -16,13 +15,13 @@ use rustc_middle::{
     mir,
     ty::{
         self,
-        layout::{LayoutCx, LayoutError, TyAndLayout},
+        layout::{LayoutCx, LayoutError, LayoutOf, TyAndLayout},
         Instance, TyCtxt,
     },
 };
 use rustc_span::def_id::DefId;
 use rustc_span::symbol::{sym, Symbol};
-use rustc_target::abi::{LayoutOf, Size};
+use rustc_target::abi::Size;
 use rustc_target::spec::abi::Abi;
 
 use crate::*;
@@ -123,13 +122,17 @@ impl fmt::Display for MiriMemoryKind {
 #[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
 pub struct Tag {
     pub alloc_id: AllocId,
-    // Stacked Borrows tag.
+    /// Stacked Borrows tag.
     pub sb: SbTag,
 }
 
 impl Provenance for Tag {
+    /// We use absolute addresses in the `offset` of a `Pointer<Tag>`.
     const OFFSET_IS_ADDR: bool = true;
 
+    /// We cannot err on partial overwrites, it happens too often in practice (due to unions).
+    const ERR_ON_PARTIAL_PTR_OVERWRITE: bool = false;
+
     fn fmt(ptr: &Pointer<Self>, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         let (tag, addr) = ptr.into_parts(); // address is absolute
         write!(f, "0x{:x}", addr.bytes())?;
@@ -190,7 +193,7 @@ impl MemoryExtra {
             Some(RefCell::new(stacked_borrows::GlobalState::new(
                 config.tracked_pointer_tag,
                 config.tracked_call_id,
-                config.track_raw,
+                config.tag_raw,
             )))
         } else {
             None
@@ -224,18 +227,22 @@ impl MemoryExtra {
     ) -> InterpResult<'tcx> {
         match this.tcx.sess.target.os.as_str() {
             "linux" => {
-                // "__cxa_thread_atexit_impl"
-                // This should be all-zero, pointer-sized.
-                let layout = this.machine.layouts.usize;
-                let place = this.allocate(layout, MiriMemoryKind::ExternStatic.into())?;
-                this.write_scalar(Scalar::from_machine_usize(0, this), &place.into())?;
-                Self::add_extern_static(this, "__cxa_thread_atexit_impl", place.ptr);
                 // "environ"
                 Self::add_extern_static(
                     this,
                     "environ",
                     this.machine.env_vars.environ.unwrap().ptr,
                 );
+                // A couple zero-initialized pointer-sized extern statics.
+                // Most of them are for weak symbols, which we all set to null (indicating that the
+                // symbol is not supported, and triggering fallback code which ends up calling a
+                // syscall that we do support).
+                for name in &["__cxa_thread_atexit_impl", "getrandom", "statx"] {
+                    let layout = this.machine.layouts.usize;
+                    let place = this.allocate(layout, MiriMemoryKind::ExternStatic.into())?;
+                    this.write_scalar(Scalar::from_machine_usize(0, this), &place.into())?;
+                    Self::add_extern_static(this, name, place.ptr);
+                }
             }
             "windows" => {
                 // "_tls_used"
@@ -300,6 +307,9 @@ pub struct Evaluator<'mir, 'tcx> {
     /// Whether to enforce the validity invariant.
     pub(crate) validate: bool,
 
+    /// Whether to enforce validity (e.g., initialization) of integers and floats.
+    pub(crate) enforce_number_validity: bool,
+
     /// Whether to enforce [ABI](Abi) of function calls.
     pub(crate) enforce_abi: bool,
 
@@ -352,6 +362,7 @@ impl<'mir, 'tcx> Evaluator<'mir, 'tcx> {
             tls: TlsData::default(),
             isolated_op: config.isolated_op,
             validate: config.validate,
+            enforce_number_validity: config.check_number_validity,
             enforce_abi: config.check_abi,
             file_handler: Default::default(),
             dir_handler: Default::default(),
@@ -422,6 +433,11 @@ impl<'mir, 'tcx> Machine<'mir, 'tcx> for Evaluator<'mir, 'tcx> {
         ecx.machine.validate
     }
 
+    #[inline(always)]
+    fn enforce_number_validity(ecx: &InterpCx<'mir, 'tcx, Self>) -> bool {
+        ecx.machine.enforce_number_validity
+    }
+
     #[inline(always)]
     fn enforce_abi(ecx: &InterpCx<'mir, 'tcx, Self>) -> bool {
         ecx.machine.enforce_abi
@@ -435,7 +451,7 @@ impl<'mir, 'tcx> Machine<'mir, 'tcx> for Evaluator<'mir, 'tcx> {
         args: &[OpTy<'tcx, Tag>],
         ret: Option<(&PlaceTy<'tcx, Tag>, mir::BasicBlock)>,
         unwind: StackPopUnwind,
-    ) -> InterpResult<'tcx, Option<&'mir mir::Body<'tcx>>> {
+    ) -> InterpResult<'tcx, Option<(&'mir mir::Body<'tcx>, ty::Instance<'tcx>)>> {
         ecx.find_mir_or_eval_fn(instance, abi, args, ret, unwind)
     }
 
@@ -453,7 +469,7 @@ impl<'mir, 'tcx> Machine<'mir, 'tcx> for Evaluator<'mir, 'tcx> {
 
     #[inline(always)]
     fn call_intrinsic(
-        ecx: &mut rustc_mir::interpret::InterpCx<'mir, 'tcx, Self>,
+        ecx: &mut rustc_const_eval::interpret::InterpCx<'mir, 'tcx, Self>,
         instance: ty::Instance<'tcx>,
         args: &[OpTy<'tcx, Tag>],
         ret: Option<(&PlaceTy<'tcx, Tag>, mir::BasicBlock)>,
@@ -478,7 +494,7 @@ impl<'mir, 'tcx> Machine<'mir, 'tcx> for Evaluator<'mir, 'tcx> {
 
     #[inline(always)]
     fn binary_ptr_op(
-        ecx: &rustc_mir::interpret::InterpCx<'mir, 'tcx, Self>,
+        ecx: &rustc_const_eval::interpret::InterpCx<'mir, 'tcx, Self>,
         bin_op: mir::BinOp,
         left: &ImmTy<'tcx, Tag>,
         right: &ImmTy<'tcx, Tag>,
@@ -486,35 +502,6 @@ impl<'mir, 'tcx> Machine<'mir, 'tcx> for Evaluator<'mir, 'tcx> {
         ecx.binary_ptr_op(bin_op, left, right)
     }
 
-    fn box_alloc(
-        ecx: &mut InterpCx<'mir, 'tcx, Self>,
-        dest: &PlaceTy<'tcx, Tag>,
-    ) -> InterpResult<'tcx> {
-        trace!("box_alloc for {:?}", dest.layout.ty);
-        let layout = ecx.layout_of(dest.layout.ty.builtin_deref(false).unwrap().ty)?;
-        // First argument: `size`.
-        // (`0` is allowed here -- this is expected to be handled by the lang item).
-        let size = Scalar::from_machine_usize(layout.size.bytes(), ecx);
-
-        // Second argument: `align`.
-        let align = Scalar::from_machine_usize(layout.align.abi.bytes(), ecx);
-
-        // Call the `exchange_malloc` lang item.
-        let malloc = ecx.tcx.lang_items().exchange_malloc_fn().unwrap();
-        let malloc = ty::Instance::mono(ecx.tcx.tcx, malloc);
-        ecx.call_function(
-            malloc,
-            Abi::Rust,
-            &[size.into(), align.into()],
-            Some(dest),
-            // Don't do anything when we are done. The `statement()` function will increment
-            // the old stack frame's stmt counter to the next statement, which means that when
-            // `exchange_malloc` returns, we go on evaluating exactly where we want to be.
-            StackPopCleanup::None { cleanup: true },
-        )?;
-        Ok(())
-    }
-
     fn thread_local_static_base_pointer(
         ecx: &mut InterpCx<'mir, 'tcx, Self>,
         def_id: DefId,
