@@ -1,11 +1,12 @@
 use log::trace;
 
 use rustc_middle::mir;
-use rustc_target::abi::{Align, LayoutOf, Size};
+use rustc_middle::ty::layout::LayoutOf;
+use rustc_span::Symbol;
+use rustc_target::abi::{Align, Size};
 use rustc_target::spec::abi::Abi;
 
 use crate::*;
-use helpers::check_arg_count;
 use shims::foreign_items::EmulateByNameResult;
 use shims::posix::fs::EvalContextExt as _;
 use shims::posix::sync::EvalContextExt as _;
@@ -15,74 +16,68 @@ impl<'mir, 'tcx: 'mir> EvalContextExt<'mir, 'tcx> for crate::MiriEvalContext<'mi
 pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx> {
     fn emulate_foreign_item_by_name(
         &mut self,
-        link_name: &str,
+        link_name: Symbol,
         abi: Abi,
         args: &[OpTy<'tcx, Tag>],
         dest: &PlaceTy<'tcx, Tag>,
         ret: mir::BasicBlock,
-    ) -> InterpResult<'tcx, EmulateByNameResult> {
+    ) -> InterpResult<'tcx, EmulateByNameResult<'mir, 'tcx>> {
         let this = self.eval_context_mut();
 
-        match link_name {
+        match &*link_name.as_str() {
             // Environment related shims
             "getenv" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref name] = check_arg_count(args)?;
+                let &[ref name] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.getenv(name)?;
-                this.write_scalar(result, dest)?;
+                this.write_pointer(result, dest)?;
             }
             "unsetenv" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref name] = check_arg_count(args)?;
+                let &[ref name] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.unsetenv(name)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
             "setenv" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref name, ref value, ref overwrite] = check_arg_count(args)?;
+                let &[ref name, ref value, ref overwrite] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 this.read_scalar(overwrite)?.to_i32()?;
                 let result = this.setenv(name, value)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
             "getcwd" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref buf, ref size] = check_arg_count(args)?;
+                let &[ref buf, ref size] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.getcwd(buf, size)?;
-                this.write_scalar(result, dest)?;
+                this.write_pointer(result, dest)?;
             }
             "chdir" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref path] = check_arg_count(args)?;
+                let &[ref path] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.chdir(path)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
 
             // File related shims
             "open" | "open64" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref path, ref flag, ref mode] = check_arg_count(args)?;
+                let &[ref path, ref flag, ref mode] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.open(path, flag, mode)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
             "fcntl" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
+                // `fcntl` is variadic. The argument count is checked based on the first argument
+                // in `this.fcntl()`, so we do not use `check_shim` here.
+                this.check_abi_and_shim_symbol_clash(abi, Abi::C { unwind: false }, link_name)?;
                 let result = this.fcntl(args)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
             "read" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref fd, ref buf, ref count] = check_arg_count(args)?;
+                let &[ref fd, ref buf, ref count] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let fd = this.read_scalar(fd)?.to_i32()?;
-                let buf = this.read_scalar(buf)?.check_init()?;
+                let buf = this.read_pointer(buf)?;
                 let count = this.read_scalar(count)?.to_machine_usize(this)?;
                 let result = this.read(fd, buf, count)?;
                 this.write_scalar(Scalar::from_machine_isize(result, this), dest)?;
             }
             "write" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref fd, ref buf, ref n] = check_arg_count(args)?;
+                let &[ref fd, ref buf, ref n] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let fd = this.read_scalar(fd)?.to_i32()?;
-                let buf = this.read_scalar(buf)?.check_init()?;
+                let buf = this.read_pointer(buf)?;
                 let count = this.read_scalar(n)?.to_machine_usize(this)?;
                 trace!("Called write({:?}, {:?}, {:?})", fd, buf, count);
                 let result = this.write(fd, buf, count)?;
@@ -90,71 +85,60 @@ pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx
                 this.write_scalar(Scalar::from_machine_isize(result, this), dest)?;
             }
             "unlink" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref path] = check_arg_count(args)?;
+                let &[ref path] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.unlink(path)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
             "symlink" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref target, ref linkpath] = check_arg_count(args)?;
+                let &[ref target, ref linkpath] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.symlink(target, linkpath)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
             "rename" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref oldpath, ref newpath] = check_arg_count(args)?;
+                let &[ref oldpath, ref newpath] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.rename(oldpath, newpath)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
             "mkdir" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref path, ref mode] = check_arg_count(args)?;
+                let &[ref path, ref mode] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.mkdir(path, mode)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
             "rmdir" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref path] = check_arg_count(args)?;
+                let &[ref path] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.rmdir(path)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
             "closedir" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref dirp] = check_arg_count(args)?;
+                let &[ref dirp] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.closedir(dirp)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
             "lseek" | "lseek64" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref fd, ref offset, ref whence] = check_arg_count(args)?;
+                let &[ref fd, ref offset, ref whence] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.lseek64(fd, offset, whence)?;
                 // "lseek" is only used on macOS which is 64bit-only, so `i64` always works.
                 this.write_scalar(Scalar::from_i64(result), dest)?;
             }
             "fsync" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref fd] = check_arg_count(args)?;
+                let &[ref fd] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.fsync(fd)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
             "fdatasync" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref fd] = check_arg_count(args)?;
+                let &[ref fd] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.fdatasync(fd)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
             "readlink" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref pathname, ref buf, ref bufsize] = check_arg_count(args)?;
+                let &[ref pathname, ref buf, ref bufsize] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.readlink(pathname, buf, bufsize)?;
                 this.write_scalar(Scalar::from_machine_isize(result, this), dest)?;
             }
 
             // Allocation
             "posix_memalign" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref ret, ref align, ref size] = check_arg_count(args)?;
+                let &[ref ret, ref align, ref size] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let ret = this.deref_operand(ret)?;
                 let align = this.read_scalar(align)?.to_machine_usize(this)?;
                 let size = this.read_scalar(size)?.to_machine_usize(this)?;
@@ -176,22 +160,21 @@ pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx
                         Size::from_bytes(size),
                         Align::from_bytes(align).unwrap(),
                         MiriMemoryKind::C.into(),
-                    );
-                    this.write_scalar(ptr, &ret.into())?;
+                    )?;
+                    this.write_pointer(ptr, &ret.into())?;
                 }
                 this.write_null(dest)?;
             }
 
             // Dynamic symbol loading
             "dlsym" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref handle, ref symbol] = check_arg_count(args)?;
+                let &[ref handle, ref symbol] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 this.read_scalar(handle)?.to_machine_usize(this)?;
-                let symbol = this.read_scalar(symbol)?.check_init()?;
+                let symbol = this.read_pointer(symbol)?;
                 let symbol_name = this.read_c_str(symbol)?;
                 if let Some(dlsym) = Dlsym::from_str(symbol_name, &this.tcx.sess.target.os)? {
                     let ptr = this.memory.create_fn_alloc(FnVal::Other(dlsym));
-                    this.write_scalar(Scalar::from(ptr), dest)?;
+                    this.write_pointer(ptr, dest)?;
                 } else {
                     this.write_null(dest)?;
                 }
@@ -199,8 +182,7 @@ pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx
 
             // Querying system information
             "sysconf" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref name] = check_arg_count(args)?;
+                let &[ref name] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let name = this.read_scalar(name)?.to_i32()?;
 
                 let sysconfs = &[
@@ -225,15 +207,15 @@ pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx
 
             // Thread-local storage
             "pthread_key_create" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref key, ref dtor] = check_arg_count(args)?;
+                let &[ref key, ref dtor] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let key_place = this.deref_operand(key)?;
-                let dtor = this.read_scalar(dtor)?.check_init()?;
+                let dtor = this.read_pointer(dtor)?;
 
                 // Extract the function type out of the signature (that seems easier than constructing it ourselves).
-                let dtor = match this.test_null(dtor)? {
-                    Some(dtor_ptr) => Some(this.memory.get_fn(dtor_ptr)?.as_instance()?),
-                    None => None,
+                let dtor = if !this.ptr_is_null(dtor)? {
+                    Some(this.memory.get_fn(dtor)?.as_instance()?)
+                } else {
+                    None
                 };
 
                 // Figure out how large a pthread TLS key actually is.
@@ -254,28 +236,25 @@ pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx
                 this.write_null(dest)?;
             }
             "pthread_key_delete" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref key] = check_arg_count(args)?;
-                let key = this.force_bits(this.read_scalar(key)?.check_init()?, key.layout.size)?;
+                let &[ref key] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
+                let key = this.read_scalar(key)?.check_init()?.to_bits(key.layout.size)?;
                 this.machine.tls.delete_tls_key(key)?;
                 // Return success (0)
                 this.write_null(dest)?;
             }
             "pthread_getspecific" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref key] = check_arg_count(args)?;
-                let key = this.force_bits(this.read_scalar(key)?.check_init()?, key.layout.size)?;
+                let &[ref key] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
+                let key = this.read_scalar(key)?.check_init()?.to_bits(key.layout.size)?;
                 let active_thread = this.get_active_thread();
                 let ptr = this.machine.tls.load_tls(key, active_thread, this)?;
                 this.write_scalar(ptr, dest)?;
             }
             "pthread_setspecific" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref key, ref new_ptr] = check_arg_count(args)?;
-                let key = this.force_bits(this.read_scalar(key)?.check_init()?, key.layout.size)?;
+                let &[ref key, ref new_ptr] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
+                let key = this.read_scalar(key)?.check_init()?.to_bits(key.layout.size)?;
                 let active_thread = this.get_active_thread();
-                let new_ptr = this.read_scalar(new_ptr)?.check_init()?;
-                this.machine.tls.store_tls(key, active_thread, this.test_null(new_ptr)?)?;
+                let new_data = this.read_scalar(new_ptr)?;
+                this.machine.tls.store_tls(key, active_thread, new_data.check_init()?, &*this.tcx)?;
 
                 // Return success (`0`).
                 this.write_null(dest)?;
@@ -283,178 +262,149 @@ pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx
 
             // Synchronization primitives
             "pthread_mutexattr_init" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref attr] = check_arg_count(args)?;
+                let &[ref attr] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.pthread_mutexattr_init(attr)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
             "pthread_mutexattr_settype" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref attr, ref kind] = check_arg_count(args)?;
+                let &[ref attr, ref kind] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.pthread_mutexattr_settype(attr, kind)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
             "pthread_mutexattr_destroy" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref attr] = check_arg_count(args)?;
+                let &[ref attr] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.pthread_mutexattr_destroy(attr)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
             "pthread_mutex_init" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref mutex, ref attr] = check_arg_count(args)?;
+                let &[ref mutex, ref attr] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.pthread_mutex_init(mutex, attr)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
             "pthread_mutex_lock" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref mutex] = check_arg_count(args)?;
+                let &[ref mutex] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.pthread_mutex_lock(mutex)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
             "pthread_mutex_trylock" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref mutex] = check_arg_count(args)?;
+                let &[ref mutex] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.pthread_mutex_trylock(mutex)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
             "pthread_mutex_unlock" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref mutex] = check_arg_count(args)?;
+                let &[ref mutex] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.pthread_mutex_unlock(mutex)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
             "pthread_mutex_destroy" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref mutex] = check_arg_count(args)?;
+                let &[ref mutex] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.pthread_mutex_destroy(mutex)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
             "pthread_rwlock_rdlock" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref rwlock] = check_arg_count(args)?;
+                let &[ref rwlock] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.pthread_rwlock_rdlock(rwlock)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
             "pthread_rwlock_tryrdlock" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref rwlock] = check_arg_count(args)?;
+                let &[ref rwlock] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.pthread_rwlock_tryrdlock(rwlock)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
             "pthread_rwlock_wrlock" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref rwlock] = check_arg_count(args)?;
+                let &[ref rwlock] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.pthread_rwlock_wrlock(rwlock)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
             "pthread_rwlock_trywrlock" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref rwlock] = check_arg_count(args)?;
+                let &[ref rwlock] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.pthread_rwlock_trywrlock(rwlock)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
             "pthread_rwlock_unlock" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref rwlock] = check_arg_count(args)?;
+                let &[ref rwlock] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.pthread_rwlock_unlock(rwlock)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
             "pthread_rwlock_destroy" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref rwlock] = check_arg_count(args)?;
+                let &[ref rwlock] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.pthread_rwlock_destroy(rwlock)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
             "pthread_condattr_init" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref attr] = check_arg_count(args)?;
+                let &[ref attr] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.pthread_condattr_init(attr)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
             "pthread_condattr_destroy" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref attr] = check_arg_count(args)?;
+                let &[ref attr] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.pthread_condattr_destroy(attr)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
             "pthread_cond_init" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref cond, ref attr] = check_arg_count(args)?;
+                let &[ref cond, ref attr] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.pthread_cond_init(cond, attr)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
             "pthread_cond_signal" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref cond] = check_arg_count(args)?;
+                let &[ref cond] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.pthread_cond_signal(cond)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
             "pthread_cond_broadcast" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref cond] = check_arg_count(args)?;
+                let &[ref cond] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.pthread_cond_broadcast(cond)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
             "pthread_cond_wait" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref cond, ref mutex] = check_arg_count(args)?;
+                let &[ref cond, ref mutex] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.pthread_cond_wait(cond, mutex)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
             "pthread_cond_timedwait" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref cond, ref mutex, ref abstime] = check_arg_count(args)?;
+                let &[ref cond, ref mutex, ref abstime] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 this.pthread_cond_timedwait(cond, mutex, abstime, dest)?;
             }
             "pthread_cond_destroy" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref cond] = check_arg_count(args)?;
+                let &[ref cond] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.pthread_cond_destroy(cond)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
 
             // Threading
             "pthread_create" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref thread, ref attr, ref start, ref arg] = check_arg_count(args)?;
+                let &[ref thread, ref attr, ref start, ref arg] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.pthread_create(thread, attr, start, arg)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
             "pthread_join" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref thread, ref retval] = check_arg_count(args)?;
+                let &[ref thread, ref retval] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.pthread_join(thread, retval)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
             "pthread_detach" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref thread] = check_arg_count(args)?;
+                let &[ref thread] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.pthread_detach(thread)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
             "pthread_self" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[] = check_arg_count(args)?;
+                let &[] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 this.pthread_self(dest)?;
             }
             "sched_yield" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[] = check_arg_count(args)?;
+                let &[] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.sched_yield()?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
             "nanosleep" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref req, ref rem] = check_arg_count(args)?;
+                let &[ref req, ref rem] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let result = this.nanosleep(req, rem)?;
                 this.write_scalar(Scalar::from_i32(result), dest)?;
             }
 
             // Miscellaneous
             "isatty" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref fd] = check_arg_count(args)?;
+                let &[ref fd] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 this.read_scalar(fd)?.to_i32()?;
                 // "returns 1 if fd is an open file descriptor referring to a terminal; otherwise 0 is returned, and errno is set to indicate the error"
                 // FIXME: we just say nothing is a terminal.
@@ -463,11 +413,10 @@ pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx
                 this.write_null(dest)?;
             }
             "pthread_atfork" => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref prepare, ref parent, ref child] = check_arg_count(args)?;
-                this.force_bits(this.read_scalar(prepare)?.check_init()?, this.memory.pointer_size())?;
-                this.force_bits(this.read_scalar(parent)?.check_init()?, this.memory.pointer_size())?;
-                this.force_bits(this.read_scalar(child)?.check_init()?, this.memory.pointer_size())?;
+                let &[ref prepare, ref parent, ref child] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
+                this.read_pointer(prepare)?;
+                this.read_pointer(parent)?;
+                this.read_pointer(child)?;
                 // We do not support forking, so there is nothing to do here.
                 this.write_null(dest)?;
             }
@@ -476,8 +425,7 @@ pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx
             // These shims are enabled only when the caller is in the standard library.
             "pthread_attr_getguardsize"
             if this.frame_in_std() => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[ref _attr, ref guard_size] = check_arg_count(args)?;
+                let &[ref _attr, ref guard_size] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 let guard_size = this.deref_operand(guard_size)?;
                 let guard_size_layout = this.libc_ty_layout("size_t")?;
                 this.write_scalar(Scalar::from_uint(crate::PAGE_SIZE, guard_size_layout.size), &guard_size.into())?;
@@ -489,29 +437,25 @@ pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx
             | "pthread_attr_init"
             | "pthread_attr_destroy"
             if this.frame_in_std() => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[_] = check_arg_count(args)?;
+                let &[_] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 this.write_null(dest)?;
             }
             | "pthread_attr_setstacksize"
             if this.frame_in_std() => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[_, _] = check_arg_count(args)?;
+                let &[_, _] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 this.write_null(dest)?;
             }
 
             | "signal"
             | "sigaltstack"
             if this.frame_in_std() => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[_, _] = check_arg_count(args)?;
+                let &[_, _] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 this.write_null(dest)?;
             }
             | "sigaction"
             | "mprotect"
             if this.frame_in_std() => {
-                this.check_abi(abi, Abi::C { unwind: false })?;
-                let &[_, _, _] = check_arg_count(args)?;
+                let &[_, _, _] = this.check_shim(abi, Abi::C { unwind: false }, link_name, args)?;
                 this.write_null(dest)?;
             }
 
