@@ -1,7 +1,7 @@
 use std::convert::TryInto;
 
 use crate::*;
-use rustc_target::abi::LayoutOf;
+use rustc_middle::ty::layout::LayoutOf;
 use rustc_target::spec::abi::Abi;
 
 impl<'mir, 'tcx> EvalContextExt<'mir, 'tcx> for crate::MiriEvalContext<'mir, 'tcx> {}
@@ -54,7 +54,7 @@ pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx
             Abi::C { unwind: false },
             &[*func_arg],
             Some(&ret_place.into()),
-            StackPopCleanup::None { cleanup: true },
+            StackPopCleanup::Root { cleanup: true },
         )?;
 
         // Restore the old active thread frame.
