@@ -38,7 +38,7 @@ pub fn main() {
 
         let j2 = spawn(move || {
             if SYNC.load(Ordering::Acquire) == 2 {
-                *c.0 //~ ERROR Data race
+                *c.0 //~ ERROR Data race detected between Read on Thread(id = 2) and Write on Thread(id = 1)
             } else {
                 0
             }
