@@ -2,19 +2,17 @@
 //! (https://github.com/rust-lang/miri/issues/1382)
 // Inlining changes the error location
 // compile-flags: -Zmir-opt-level=0
-#![feature(const_panic)]
 #![feature(never_type)]
 #![warn(warnings, const_err)]
 
 struct PrintName<T>(T);
 impl<T> PrintName<T> {
-    const VOID: ! = panic!(); //~WARN any use of this value will cause an error
-    //~^ WARN this was previously accepted
+    const VOID: ! = panic!(); //~ERROR evaluation of `PrintName::<i32>::VOID` failed
 }
 
 fn no_codegen<T>() {
     if false {
-        let _ = PrintName::<T>::VOID; //~ERROR referenced constant has errors
+        let _ = PrintName::<T>::VOID; //~ERROR error occurred: encountered constant
     }
 }
 fn main() {
