@@ -1,5 +1,3 @@
-#![allow(unused_variables)]
-
 mod safe {
     use std::slice::from_raw_parts_mut;
 
@@ -13,7 +11,7 @@ mod safe {
 fn main() {
     let v = vec![0,1,2];
     let v1 = safe::as_mut_slice(&v);
-    let v2 = safe::as_mut_slice(&v);
-    v1[1] = 5; //~ ERROR does not exist on the stack
-    v1[1] = 6;
+    let _v2 = safe::as_mut_slice(&v);
+    v1[1] = 5;
+    //~^ ERROR borrow stack
 }
