@@ -1,5 +1,6 @@
-use std::{future::Future, pin::Pin, task::Poll, ptr};
-use std::task::{Waker, RawWaker, RawWakerVTable, Context};
+#![feature(never_type)]
+
+use std::future::Future;
 
 // See if we can run a basic `async fn`
 pub async fn foo(x: &u32, y: u32) -> u32 {
@@ -44,37 +45,46 @@ async fn partial_init(x: u32) -> u32 {
     let _x: (String, !) = (String::new(), return async { x + x }.await);
 }
 
-fn run_fut(mut fut: impl Future<Output=u32>, output: u32) {
-    fn raw_waker_clone(_this: *const ()) -> RawWaker {
-        panic!("unimplemented");
-    }
-    fn raw_waker_wake(_this: *const ()) {
-        panic!("unimplemented");
-    }
-    fn raw_waker_wake_by_ref(_this: *const ()) {
-        panic!("unimplemented");
-    }
-    fn raw_waker_drop(_this: *const ()) {}
+async fn read_exact(_from: &mut &[u8], _to: &mut [u8]) -> Option<()> {
+    Some(())
+}
 
-    static RAW_WAKER: RawWakerVTable = RawWakerVTable::new(
-        raw_waker_clone,
-        raw_waker_wake,
-        raw_waker_wake_by_ref,
-        raw_waker_drop,
-    );
+async fn hello_world() {
+    let data = [0u8; 1];
+    let mut reader = &data[..];
 
-    let waker = unsafe { Waker::from_raw(RawWaker::new(ptr::null(), &RAW_WAKER)) };
+    let mut marker = [0u8; 1];
+    read_exact(&mut reader, &mut marker).await.unwrap();
+}
+
+fn run_fut<T>(fut: impl Future<Output = T>) -> T {
+    use std::sync::Arc;
+    use std::task::{Context, Poll, Wake, Waker};
+
+    struct MyWaker;
+    impl Wake for MyWaker {
+        fn wake(self: Arc<Self>) {
+            unimplemented!()
+        }
+    }
+
+    let waker = Waker::from(Arc::new(MyWaker));
     let mut context = Context::from_waker(&waker);
-    assert_eq!(unsafe { Pin::new_unchecked(&mut fut) }.poll(&mut context), Poll::Ready(output));
+
+    let mut pinned = Box::pin(fut);
+    loop {
+        match pinned.as_mut().poll(&mut context) {
+            Poll::Pending => continue,
+            Poll::Ready(v) => return v,
+        }
+    }
 }
 
 fn main() {
     let x = 5;
-    run_fut(foo(&x, 7), 31);
-
-    run_fut(build_aggregate(1, 2, 3, 4), 10);
-
-    run_fut(includes_never(false, 4), 16);
-
-    run_fut(partial_init(4), 8);
+    assert_eq!(run_fut(foo(&x, 7)), 31);
+    assert_eq!(run_fut(build_aggregate(1, 2, 3, 4)), 10);
+    assert_eq!(run_fut(includes_never(false, 4)), 16);
+    assert_eq!(run_fut(partial_init(4)), 8);
+    run_fut(hello_world());
 }
