@@ -1,24 +1,12 @@
-// Copyright 2016 The RLS Project Developers.
-//
-// Licensed under the Apache License, Version 2.0 <LICENSE-APACHE or
-// http://www.apache.org/licenses/LICENSE-2.0> or the MIT license
-// <LICENSE-MIT or http://opensource.org/licenses/MIT>, at your
-// option. This file may not be copied, modified, or distributed
-// except according to those terms.
-
-#![feature(type_ascription)]
-#![feature(crate_visibility_modifier)]
+#![warn(rust_2018_idioms)]
 
 #[macro_use]
 extern crate derive_new;
 #[macro_use]
 extern crate log;
-extern crate fst;
-extern crate itertools;
-extern crate json;
+
 extern crate rls_data as data;
 extern crate rls_span as span;
-extern crate rustc_serialize;
 
 mod analysis;
 mod listings;
@@ -31,18 +19,21 @@ mod test;
 mod util;
 
 use analysis::Analysis;
-pub use analysis::{Def, Ref};
+pub use analysis::{Def, Ident, IdentKind, Ref};
 pub use loader::{AnalysisLoader, CargoAnalysisLoader, SearchDirectory, Target};
-pub use raw::{name_space_for_def_kind, read_analysis_from_files, CrateId, DefKind};
+pub use raw::{
+    deserialize_crate_data, name_space_for_def_kind, read_analysis_from_files, read_crate_data,
+    Crate, CrateId, DefKind,
+};
 pub use symbol_query::SymbolQuery;
 
 use std::collections::HashMap;
+use std::fmt::Debug;
 use std::path::{Path, PathBuf};
 use std::sync::Mutex;
 use std::time::{Instant, SystemTime};
 use std::u64;
 
-#[derive(Debug)]
 pub struct AnalysisHost<L: AnalysisLoader = CargoAnalysisLoader> {
     analysis: Mutex<Option<Analysis>>,
     master_crate_map: Mutex<HashMap<CrateId, u32>>,
@@ -90,15 +81,13 @@ impl Id {
     fn from_crate_and_local(crate_id: u32, local_id: u32) -> Id {
         // Use global crate number for high order bits,
         // then index for least significant bits.
-        Id(((crate_id as u64) << 32) | (local_id as u64))
+        Id((u64::from(crate_id) << 32) | u64::from(local_id))
     }
 }
 
 /// Used to indicate a missing index in the Id.
 pub const NULL: Id = Id(u64::MAX);
 
-type Blacklist<'a> = &'a [&'static str];
-
 macro_rules! clone_field {
     ($field: ident) => {
         |x| x.$field.clone()
@@ -138,7 +127,7 @@ impl<L: AnalysisLoader> AnalysisHost<L> {
         analysis: Vec<data::Analysis>,
         path_prefix: &Path,
         base_dir: &Path,
-        blacklist: Blacklist,
+        blacklist: &[impl AsRef<str> + Debug],
     ) -> AResult<()> {
         self.reload_with_blacklist(path_prefix, base_dir, blacklist)?;
 
@@ -155,14 +144,14 @@ impl<L: AnalysisLoader> AnalysisHost<L> {
     }
 
     pub fn reload(&self, path_prefix: &Path, base_dir: &Path) -> AResult<()> {
-        self.reload_with_blacklist(path_prefix, base_dir, &[])
+        self.reload_with_blacklist(path_prefix, base_dir, &[] as &[&str])
     }
 
     pub fn reload_with_blacklist(
         &self,
         path_prefix: &Path,
         base_dir: &Path,
-        blacklist: Blacklist,
+        blacklist: &[impl AsRef<str> + Debug],
     ) -> AResult<()> {
         trace!("reload_with_blacklist {:?} {:?} {:?}", path_prefix, base_dir, blacklist);
         let empty = self.analysis.lock()?.is_none();
@@ -185,14 +174,14 @@ impl<L: AnalysisLoader> AnalysisHost<L> {
 
     /// Reloads the entire project's analysis data.
     pub fn hard_reload(&self, path_prefix: &Path, base_dir: &Path) -> AResult<()> {
-        self.hard_reload_with_blacklist(path_prefix, base_dir, &[])
+        self.hard_reload_with_blacklist(path_prefix, base_dir, &[] as &[&str])
     }
 
     pub fn hard_reload_with_blacklist(
         &self,
         path_prefix: &Path,
         base_dir: &Path,
-        blacklist: Blacklist,
+        blacklist: &[impl AsRef<str> + Debug],
     ) -> AResult<()> {
         trace!("hard_reload {:?} {:?}", path_prefix, base_dir);
         // We're going to create a dummy AnalysisHost that we will fill with data,
@@ -242,7 +231,7 @@ impl<L: AnalysisLoader> AnalysisHost<L> {
     }
 
     pub fn get_def(&self, id: Id) -> AResult<Def> {
-        self.with_analysis(|a| a.with_defs(id, |def| def.clone()))
+        self.with_analysis(|a| a.with_defs(id, Clone::clone))
     }
 
     pub fn goto_def(&self, span: &Span) -> AResult<Span> {
@@ -338,7 +327,7 @@ impl<L: AnalysisLoader> AnalysisHost<L> {
         let time = t_start.elapsed();
         info!(
             "find_all_refs: {}s",
-            time.as_secs() as f64 + time.subsec_nanos() as f64 / 1_000_000_000.0
+            time.as_secs() as f64 + f64::from(time.subsec_nanos()) / 1_000_000_000.0
         );
         result
     }
@@ -373,7 +362,7 @@ impl<L: AnalysisLoader> AnalysisHost<L> {
         let time = t_start.elapsed();
         info!(
             "query_defs: {}",
-            time.as_secs() as f64 + time.subsec_nanos() as f64 / 1_000_000_000.0
+            time.as_secs() as f64 + f64::from(time.subsec_nanos()) / 1_000_000_000.0
         );
 
         result
@@ -386,7 +375,7 @@ impl<L: AnalysisLoader> AnalysisHost<L> {
         let result = self.with_analysis(|a| {
             Some(a.with_def_names(name, |defs| {
                 info!("defs: {:?}", defs);
-                defs.into_iter()
+                defs.iter()
                     .flat_map(|id| {
                         a.with_ref_spans(*id, |refs| {
                             Some(
@@ -400,12 +389,15 @@ impl<L: AnalysisLoader> AnalysisHost<L> {
                         .unwrap_or_else(Vec::new)
                         .into_iter()
                     })
-                    .collect(): Vec<Span>
+                    .collect::<Vec<Span>>()
             }))
         });
 
         let time = t_start.elapsed();
-        info!("search: {}s", time.as_secs() as f64 + time.subsec_nanos() as f64 / 1_000_000_000.0);
+        info!(
+            "search: {}s",
+            time.as_secs() as f64 + f64::from(time.subsec_nanos()) / 1_000_000_000.0
+        );
         result
     }
 
@@ -423,7 +415,7 @@ impl<L: AnalysisLoader> AnalysisHost<L> {
         let time = t_start.elapsed();
         info!(
             "find_all_refs_by_id: {}s",
-            time.as_secs() as f64 + time.subsec_nanos() as f64 / 1_000_000_000.0
+            time.as_secs() as f64 + f64::from(time.subsec_nanos()) / 1_000_000_000.0
         );
         result
     }
@@ -434,7 +426,13 @@ impl<L: AnalysisLoader> AnalysisHost<L> {
 
     /// Search for a symbol name, returning a list of def_ids for that name.
     pub fn search_for_id(&self, name: &str) -> AResult<Vec<Id>> {
-        self.with_analysis(|a| Some(a.with_def_names(name, |defs| defs.clone())))
+        self.with_analysis(|a| Some(a.with_def_names(name, Clone::clone)))
+    }
+
+    /// Returns all identifiers which overlap the given span.
+    #[cfg(feature = "idents")]
+    pub fn idents(&self, span: &Span) -> AResult<Vec<Ident>> {
+        self.with_analysis(|a| Some(a.idents(span)))
     }
 
     pub fn symbols(&self, file_name: &Path) -> AResult<Vec<SymbolResult>> {
@@ -550,23 +548,20 @@ impl<L: AnalysisLoader> AnalysisHost<L> {
 }
 
 impl ::std::fmt::Display for Id {
-    fn fmt(&self, f: &mut ::std::fmt::Formatter) -> ::std::fmt::Result {
-        self.0.fmt(f)
+    fn fmt(&self, f: &mut ::std::fmt::Formatter<'_>) -> ::std::fmt::Result {
+        ::std::fmt::Display::fmt(&self.0, f)
     }
 }
 
-impl ::std::error::Error for AError {
-    fn description(&self) -> &str {
-        match *self {
-            AError::MutexPoison => "poison error in a mutex (usually a secondary error)",
-            AError::Unclassified => "unknown error",
-        }
-    }
-}
+impl ::std::error::Error for AError {}
 
 impl ::std::fmt::Display for AError {
-    fn fmt(&self, f: &mut ::std::fmt::Formatter) -> ::std::fmt::Result {
-        write!(f, "{}", ::std::error::Error::description(self))
+    fn fmt(&self, f: &mut ::std::fmt::Formatter<'_>) -> ::std::fmt::Result {
+        let description = match self {
+            AError::MutexPoison => "poison error in a mutex (usually a secondary error)",
+            AError::Unclassified => "unknown error",
+        };
+        write!(f, "{}", description)
     }
 }
 
