@@ -15,7 +15,6 @@ use lsp_types::{
 };
 use rls_span::compiler::DiagnosticSpan;
 use serde_derive::Deserialize;
-use serde_json;
 use url::Url;
 
 pub use lsp_types::Diagnostic;
@@ -176,7 +175,11 @@ fn format_notes(children: &[AssociatedMessage], primary: &DiagnosticSpan) -> Opt
         } else if spans.len() == 1 && spans[0].is_within(primary) {
             add_message_to_notes!(message);
             if let Some(ref suggested) = spans[0].suggested_replacement {
-                notes.push_str(&format!(": `{}`", suggested));
+                if !suggested.is_empty() {
+                    // Only show the suggestion when it is non-empty.
+                    // This matches rustc's behavior.
+                    notes.push_str(&format!(": `{}`", suggested));
+                }
             }
         }
     }
@@ -598,12 +601,6 @@ help: consider borrowing here: `&string`"#,
         assert!(messages[0].1.is_empty(), "{:?}", messages[0].1);
     }
 
-    /// ```
-    /// fn main() {
-    ///     let mut out = String::new();
-    ///     write!(out, "{}", 123);
-    /// }
-    /// ```
     #[test]
     fn macro_error_no_trait() {
         let diag = parse_compiler_message(
