@@ -679,7 +679,9 @@ fn client_dependency_typo_and_fix() {
     let diag = rls.wait_for_diagnostics();
     assert_eq!(diag.diagnostics.len(), 1);
     assert_eq!(diag.diagnostics[0].severity, Some(DiagnosticSeverity::Error));
-    assert!(diag.diagnostics[0].message.contains("no matching package named `auto-cfg`"));
+    assert!(diag.diagnostics[0]
+        .message
+        .contains("no matching package found\nsearched package name: `auto-cfg`"));
 
     let change_manifest = |contents: &str| {
         std::fs::write(root_path.join("Cargo.toml"), contents).unwrap();
@@ -2147,7 +2149,7 @@ fn client_fail_uninitialized_request() {
         },
     );
 
-    rls.block_on(async { tokio::time::delay_for(Duration::from_secs(1)).await }).unwrap();
+    rls.block_on(async { tokio::time::sleep(Duration::from_secs(1)).await }).unwrap();
 
     let err = jsonrpc_core::Failure::deserialize(rls.messages().last().unwrap()).unwrap();
     assert_eq!(err.id, jsonrpc_core::Id::Num(ID));
