@@ -70,13 +70,13 @@ extern "C" {
             #[on_extern_block]  // AttrId { is_doc_comment: false, ast_index: 0 }
             extern "C" {
                 #[on_extern_type]  // AttrId { is_doc_comment: false, ast_index: 0 }
-                pub(self) type ExType;  // extern
+                pub(self) type ExType;
 
                 #[on_extern_static]  // AttrId { is_doc_comment: false, ast_index: 0 }
-                pub(self) static EX_STATIC: u8 = _;  // extern
+                pub(self) static EX_STATIC: u8 = _;
 
                 #[on_extern_fn]  // AttrId { is_doc_comment: false, ast_index: 0 }
-                // flags = 0x60
+                // flags = 0x20
                 pub(self) fn ex_fn() -> ();
             }
         "##]],
@@ -176,7 +176,7 @@ trait Tr: SuperTrait + 'lifetime {
             // flags = 0x2
             pub(self) fn f(
                 #[attr]  // AttrId { is_doc_comment: false, ast_index: 0 }
-                _: u8,
+                arg: u8,
                 _: (),
             ) -> ();
 
@@ -282,6 +282,7 @@ struct S {
     a: Mixed<'a, T, Item=(), OtherItem=u8>,
     b: <Fully as Qualified>::Syntax,
     c: <TypeAnchored>::Path::<'a>,
+    d: dyn for<'a> Trait<'a>,
 }
         "#,
         expect![[r#"
@@ -289,6 +290,7 @@ struct S {
                 pub(self) a: Mixed<'a, T, Item = (), OtherItem = u8>,
                 pub(self) b: Qualified<Self=Fully>::Syntax,
                 pub(self) c: <TypeAnchored>::Path<'a>,
+                pub(self) d: dyn for<'a> Trait<'a>,
             }
         "#]],
     )
@@ -302,7 +304,7 @@ struct S<'a, 'b: 'a, T: Copy + 'a + 'b, const K: u8 = 0> {
     field: &'a &'b T,
 }
 
-struct Tuple<T: Copy>(T);
+struct Tuple<T: Copy, U: ?Sized>(T, U);
 
 impl<'a, 'b: 'a, T: Copy + 'a + 'b, const K: u8 = 0> S<'a, 'b, T, K> {
     fn f<G: 'a>(arg: impl Copy) -> impl Copy {}
@@ -311,7 +313,7 @@ impl<'a, 'b: 'a, T: Copy + 'a + 'b, const K: u8 = 0> S<'a, 'b, T, K> {
 enum Enum<'a, T, const U: u8> {}
 union Union<'a, T, const U: u8> {}
 
-trait Tr<'a, T: 'a>: Super {}
+trait Tr<'a, T: 'a>: Super where Self: for<'a> Tr<'a, T> {}
         "#,
         expect![[r#"
             pub(self) struct S<'a, 'b, T, const K: u8>
@@ -323,11 +325,13 @@ trait Tr<'a, T: 'a>: Super {}
                 pub(self) field: &'a &'b T,
             }
 
-            pub(self) struct Tuple<T>(
+            pub(self) struct Tuple<T, U>(
                 pub(self) 0: T,
+                pub(self) 1: U,
             )
             where
-                T: Copy;
+                T: Copy,
+                U: ?Sized;
 
             impl<'a, 'b, T, const K: u8> S<'a, 'b, T, K>
             where
@@ -336,12 +340,11 @@ trait Tr<'a, T: 'a>: Super {}
                 T: 'b
             {
                 // flags = 0x2
-                pub(self) fn f<G, _anon_1>(
-                    _: impl Copy,
+                pub(self) fn f<G>(
+                    arg: impl Copy,
                 ) -> impl Copy
                 where
-                    G: 'a,
-                    _anon_1: Copy;
+                    G: 'a;
             }
 
             pub(self) enum Enum<'a, T, const U: u8> {
@@ -353,7 +356,8 @@ trait Tr<'a, T: 'a>: Super {}
             pub(self) trait Tr<'a, Self, T>
             where
                 Self: Super,
-                T: 'a
+                T: 'a,
+                Self: for<'a> Tr<'a, T>
             {
             }
         "#]],
