@@ -11,7 +11,6 @@
 //! `ReachedFixedPoint` signals about this.
 
 use base_db::Edition;
-use hir_expand::name;
 use hir_expand::name::Name;
 
 use crate::{
@@ -55,17 +54,26 @@ impl ResolvePathResult {
         segment_index: Option<usize>,
         krate: Option<CrateId>,
     ) -> ResolvePathResult {
-        ResolvePathResult { resolved_def, reached_fixedpoint, segment_index, krate }
+        ResolvePathResult { resolved_def, segment_index, reached_fixedpoint, krate }
     }
 }
 
 impl DefMap {
-    pub(super) fn resolve_name_in_extern_prelude(&self, name: &Name) -> PerNs {
-        if name == &name!(self) {
-            cov_mark::hit!(extern_crate_self_as);
-            return PerNs::types(self.module_id(self.root).into(), Visibility::Public);
-        }
-        self.extern_prelude
+    pub(super) fn resolve_name_in_extern_prelude(
+        &self,
+        db: &dyn DefDatabase,
+        name: &Name,
+    ) -> PerNs {
+        let arc;
+        let root = match self.block {
+            Some(_) => {
+                arc = self.crate_root(db).def_map(db);
+                &*arc
+            }
+            None => self,
+        };
+
+        root.extern_prelude
             .get(name)
             .map_or(PerNs::none(), |&it| PerNs::types(it, Visibility::Public))
     }
@@ -79,7 +87,7 @@ impl DefMap {
         let mut vis = match visibility {
             RawVisibility::Module(path) => {
                 let (result, remaining) =
-                    self.resolve_path(db, original_module, &path, BuiltinShadowMode::Module);
+                    self.resolve_path(db, original_module, path, BuiltinShadowMode::Module);
                 if remaining.is_some() {
                     return None;
                 }
@@ -102,7 +110,7 @@ impl DefMap {
             if self.block_id() != m.block {
                 cov_mark::hit!(adjust_vis_in_block_def_map);
                 vis = Visibility::Module(self.module_id(self.root()));
-                log::debug!("visibility {:?} points outside DefMap, adjusting to {:?}", m, vis);
+                tracing::debug!("visibility {:?} points outside DefMap, adjusting to {:?}", m, vis);
             }
         }
 
@@ -156,7 +164,7 @@ impl DefMap {
         }
     }
 
-    fn resolve_path_fp_with_macro_single(
+    pub(super) fn resolve_path_fp_with_macro_single(
         &self,
         db: &dyn DefDatabase,
         mode: ResolveMode,
@@ -164,6 +172,12 @@ impl DefMap {
         path: &ModPath,
         shadow: BuiltinShadowMode,
     ) -> ResolvePathResult {
+        let graph = db.crate_graph();
+        let _cx = stdx::panic_context::enter(format!(
+            "DefMap {:?} crate_name={:?} block={:?} path={}",
+            self.krate, graph[self.krate].display_name, self.block, path
+        ));
+
         let mut segments = path.segments().iter().enumerate();
         let mut curr_per_ns: PerNs = match path.kind {
             PathKind::DollarCrate(krate) => {
@@ -190,8 +204,8 @@ impl DefMap {
                     Some((idx, segment)) => (idx, segment),
                     None => return ResolvePathResult::empty(ReachedFixedPoint::Yes),
                 };
-                log::debug!("resolving {:?} in crate root (+ extern prelude)", segment);
-                self.resolve_name_in_crate_root_or_extern_prelude(&segment)
+                tracing::debug!("resolving {:?} in crate root (+ extern prelude)", segment);
+                self.resolve_name_in_crate_root_or_extern_prelude(db, segment)
             }
             PathKind::Plain => {
                 let (_, segment) = match segments.next() {
@@ -207,8 +221,8 @@ impl DefMap {
                 let prefer_module =
                     if path.segments().len() == 1 { shadow } else { BuiltinShadowMode::Module };
 
-                log::debug!("resolving {:?} in module", segment);
-                self.resolve_name_in_module(db, original_module, &segment, prefer_module)
+                tracing::debug!("resolving {:?} in module", segment);
+                self.resolve_name_in_module(db, original_module, segment, prefer_module)
             }
             PathKind::Super(lvl) => {
                 let mut module = original_module;
@@ -222,7 +236,11 @@ impl DefMap {
                                     PathKind::Super(lvl - i),
                                     path.segments().to_vec(),
                                 );
-                                log::debug!("`super` path: {} -> {} in parent map", path, new_path);
+                                tracing::debug!(
+                                    "`super` path: {} -> {} in parent map",
+                                    path,
+                                    new_path
+                                );
                                 return block.parent.def_map(db).resolve_path_fp_with_macro(
                                     db,
                                     mode,
@@ -232,7 +250,7 @@ impl DefMap {
                                 );
                             }
                             None => {
-                                log::debug!("super path in root module");
+                                tracing::debug!("super path in root module");
                                 return ResolvePathResult::empty(ReachedFixedPoint::Yes);
                             }
                         },
@@ -255,8 +273,8 @@ impl DefMap {
                     Some((_, segment)) => segment,
                     None => return ResolvePathResult::empty(ReachedFixedPoint::Yes),
                 };
-                if let Some(def) = self.extern_prelude.get(&segment) {
-                    log::debug!("absolute path {:?} resolved to crate {:?}", path, def);
+                if let Some(def) = self.extern_prelude.get(segment) {
+                    tracing::debug!("absolute path {:?} resolved to crate {:?}", path, def);
                     PerNs::types(*def, Visibility::Public)
                 } else {
                     return ResolvePathResult::empty(ReachedFixedPoint::No); // extern crate declarations can add to the extern prelude
@@ -285,7 +303,7 @@ impl DefMap {
                             PathKind::Super(0),
                             path.segments()[i..].iter().cloned(),
                         );
-                        log::debug!("resolving {:?} in other crate", path);
+                        tracing::debug!("resolving {:?} in other crate", path);
                         let defp_map = module.def_map(db);
                         let (def, s) = defp_map.resolve_path(db, module.local_id, &path, shadow);
                         return ResolvePathResult::with(
@@ -305,13 +323,13 @@ impl DefMap {
                     };
 
                     // Since it is a qualified path here, it should not contains legacy macros
-                    module_data.scope.get(&segment)
+                    module_data.scope.get(segment)
                 }
                 ModuleDefId::AdtId(AdtId::EnumId(e)) => {
                     // enum variant
                     cov_mark::hit!(can_import_enum_variant);
                     let enum_data = db.enum_data(e);
-                    match enum_data.variant(&segment) {
+                    match enum_data.variant(segment) {
                         Some(local_id) => {
                             let variant = EnumVariantId { parent: e, local_id };
                             match &*enum_data.variants[local_id].variant_data {
@@ -337,7 +355,7 @@ impl DefMap {
                 s => {
                     // could be an inherent method call in UFCS form
                     // (`Struct::method`), or some other kind of associated item
-                    log::debug!(
+                    tracing::debug!(
                         "path segment {:?} resolved to non-module {:?}, but is not last",
                         segment,
                         curr,
@@ -373,35 +391,45 @@ impl DefMap {
             .get_legacy_macro(name)
             .map_or_else(PerNs::none, |m| PerNs::macros(m, Visibility::Public));
         let from_scope = self[module].scope.get(name);
-        let from_builtin = BUILTIN_SCOPE.get(name).copied().unwrap_or_else(PerNs::none);
+        let from_builtin = match self.block {
+            Some(_) => {
+                // Only resolve to builtins in the root `DefMap`.
+                PerNs::none()
+            }
+            None => BUILTIN_SCOPE.get(name).copied().unwrap_or_else(PerNs::none),
+        };
         let from_scope_or_builtin = match shadow {
             BuiltinShadowMode::Module => from_scope.or(from_builtin),
-            BuiltinShadowMode::Other => {
-                if let Some(ModuleDefId::ModuleId(_)) = from_scope.take_types() {
-                    from_builtin.or(from_scope)
-                } else {
-                    from_scope.or(from_builtin)
-                }
-            }
-        };
-        // Give precedence to names in outer `DefMap`s over the extern prelude; only check prelude
-        // from the crate DefMap.
-        let from_extern_prelude = match self.block {
-            Some(_) => PerNs::none(),
-            None => self
-                .extern_prelude
-                .get(name)
-                .map_or(PerNs::none(), |&it| PerNs::types(it, Visibility::Public)),
+            BuiltinShadowMode::Other => match from_scope.take_types() {
+                Some(ModuleDefId::ModuleId(_)) => from_builtin.or(from_scope),
+                Some(_) | None => from_scope.or(from_builtin),
+            },
         };
+        let from_extern_prelude = self
+            .extern_prelude
+            .get(name)
+            .map_or(PerNs::none(), |&it| PerNs::types(it, Visibility::Public));
 
         let from_prelude = self.resolve_in_prelude(db, name);
 
         from_legacy_macro.or(from_scope_or_builtin).or(from_extern_prelude).or(from_prelude)
     }
 
-    fn resolve_name_in_crate_root_or_extern_prelude(&self, name: &Name) -> PerNs {
-        let from_crate_root = self[self.root].scope.get(name);
-        let from_extern_prelude = self.resolve_name_in_extern_prelude(name);
+    fn resolve_name_in_crate_root_or_extern_prelude(
+        &self,
+        db: &dyn DefDatabase,
+        name: &Name,
+    ) -> PerNs {
+        let arc;
+        let crate_def_map = match self.block {
+            Some(_) => {
+                arc = self.crate_root(db).def_map(db);
+                &arc
+            }
+            None => self,
+        };
+        let from_crate_root = crate_def_map[crate_def_map.root].scope.get(name);
+        let from_extern_prelude = self.resolve_name_in_extern_prelude(db, name);
 
         from_crate_root.or(from_extern_prelude)
     }
