@@ -4,8 +4,9 @@ use chalk_ir::{FloatTy, IntTy, Mutability, Scalar, UintTy};
 use hir_def::{
     builtin_type::{BuiltinFloat, BuiltinInt, BuiltinType, BuiltinUint},
     type_ref::Rawness,
-    AssocContainerId, FunctionId, GenericDefId, HasModule, Lookup, TraitId,
+    FunctionId, GenericDefId, HasModule, ItemContainerId, Lookup, TraitId,
 };
+use syntax::SmolStr;
 
 use crate::{
     db::HirDatabase, from_assoc_type_id, from_chalk_trait_id, from_foreign_def_id,
@@ -45,30 +46,30 @@ pub trait TyExt {
 
 impl TyExt for Ty {
     fn is_unit(&self) -> bool {
-        matches!(self.kind(&Interner), TyKind::Tuple(0, _))
+        matches!(self.kind(Interner), TyKind::Tuple(0, _))
     }
 
     fn is_never(&self) -> bool {
-        matches!(self.kind(&Interner), TyKind::Never)
+        matches!(self.kind(Interner), TyKind::Never)
     }
 
     fn is_unknown(&self) -> bool {
-        matches!(self.kind(&Interner), TyKind::Error)
+        matches!(self.kind(Interner), TyKind::Error)
     }
 
     fn is_ty_var(&self) -> bool {
-        matches!(self.kind(&Interner), TyKind::InferenceVar(_, _))
+        matches!(self.kind(Interner), TyKind::InferenceVar(_, _))
     }
 
     fn as_adt(&self) -> Option<(hir_def::AdtId, &Substitution)> {
-        match self.kind(&Interner) {
+        match self.kind(Interner) {
             TyKind::Adt(AdtId(adt), parameters) => Some((*adt, parameters)),
             _ => None,
         }
     }
 
     fn as_builtin(&self) -> Option<BuiltinType> {
-        match self.kind(&Interner) {
+        match self.kind(Interner) {
             TyKind::Str => Some(BuiltinType::Str),
             TyKind::Scalar(Scalar::Bool) => Some(BuiltinType::Bool),
             TyKind::Scalar(Scalar::Char) => Some(BuiltinType::Char),
@@ -97,28 +98,27 @@ impl TyExt for Ty {
     }
 
     fn as_tuple(&self) -> Option<&Substitution> {
-        match self.kind(&Interner) {
+        match self.kind(Interner) {
             TyKind::Tuple(_, substs) => Some(substs),
             _ => None,
         }
     }
 
     fn as_fn_def(&self, db: &dyn HirDatabase) -> Option<FunctionId> {
-        if let Some(CallableDefId::FunctionId(func)) = self.callable_def(db) {
-            Some(func)
-        } else {
-            None
+        match self.callable_def(db) {
+            Some(CallableDefId::FunctionId(func)) => Some(func),
+            Some(CallableDefId::StructId(_) | CallableDefId::EnumVariantId(_)) | None => None,
         }
     }
     fn as_reference(&self) -> Option<(&Ty, Lifetime, Mutability)> {
-        match self.kind(&Interner) {
+        match self.kind(Interner) {
             TyKind::Ref(mutability, lifetime, ty) => Some((ty, lifetime.clone(), *mutability)),
             _ => None,
         }
     }
 
     fn as_reference_or_ptr(&self) -> Option<(&Ty, Rawness, Mutability)> {
-        match self.kind(&Interner) {
+        match self.kind(Interner) {
             TyKind::Ref(mutability, _, ty) => Some((ty, Rawness::Ref, *mutability)),
             TyKind::Raw(mutability, ty) => Some((ty, Rawness::RawPtr, *mutability)),
             _ => None,
@@ -126,7 +126,7 @@ impl TyExt for Ty {
     }
 
     fn as_generic_def(&self, db: &dyn HirDatabase) -> Option<GenericDefId> {
-        match *self.kind(&Interner) {
+        match *self.kind(Interner) {
             TyKind::Adt(AdtId(adt), ..) => Some(adt.into()),
             TyKind::FnDef(callable, ..) => {
                 Some(db.lookup_intern_callable_def(callable.into()).into())
@@ -138,22 +138,22 @@ impl TyExt for Ty {
     }
 
     fn callable_def(&self, db: &dyn HirDatabase) -> Option<CallableDefId> {
-        match self.kind(&Interner) {
+        match self.kind(Interner) {
             &TyKind::FnDef(def, ..) => Some(db.lookup_intern_callable_def(def.into())),
             _ => None,
         }
     }
 
     fn callable_sig(&self, db: &dyn HirDatabase) -> Option<CallableSig> {
-        match self.kind(&Interner) {
+        match self.kind(Interner) {
             TyKind::Function(fn_ptr) => Some(CallableSig::from_fn_ptr(fn_ptr)),
             TyKind::FnDef(def, parameters) => {
                 let callable_def = db.lookup_intern_callable_def((*def).into());
                 let sig = db.callable_item_signature(callable_def);
-                Some(sig.substitute(&Interner, &parameters))
+                Some(sig.substitute(Interner, &parameters))
             }
             TyKind::Closure(.., substs) => {
-                let sig_param = substs.at(&Interner, 0).assert_ty_ref(&Interner);
+                let sig_param = substs.at(Interner, 0).assert_ty_ref(Interner);
                 sig_param.callable_sig(db)
             }
             _ => None,
@@ -161,7 +161,7 @@ impl TyExt for Ty {
     }
 
     fn dyn_trait(&self) -> Option<TraitId> {
-        let trait_ref = match self.kind(&Interner) {
+        let trait_ref = match self.kind(Interner) {
             TyKind::Dyn(dyn_ty) => dyn_ty.bounds.skip_binders().interned().get(0).and_then(|b| {
                 match b.skip_binders() {
                     WhereClause::Implemented(trait_ref) => Some(trait_ref),
@@ -175,30 +175,30 @@ impl TyExt for Ty {
 
     fn strip_references(&self) -> &Ty {
         let mut t: &Ty = self;
-        while let TyKind::Ref(_mutability, _lifetime, ty) = t.kind(&Interner) {
+        while let TyKind::Ref(_mutability, _lifetime, ty) = t.kind(Interner) {
             t = ty;
         }
         t
     }
 
     fn impl_trait_bounds(&self, db: &dyn HirDatabase) -> Option<Vec<QuantifiedWhereClause>> {
-        match self.kind(&Interner) {
-            TyKind::OpaqueType(opaque_ty_id, ..) => {
+        match self.kind(Interner) {
+            TyKind::OpaqueType(opaque_ty_id, subst) => {
                 match db.lookup_intern_impl_trait_id((*opaque_ty_id).into()) {
                     ImplTraitId::AsyncBlockTypeImplTrait(def, _expr) => {
                         let krate = def.module(db.upcast()).krate();
                         if let Some(future_trait) = db
-                            .lang_item(krate, "future_trait".into())
+                            .lang_item(krate, SmolStr::new_inline("future_trait"))
                             .and_then(|item| item.as_trait())
                         {
                             // This is only used by type walking.
                             // Parameters will be walked outside, and projection predicate is not used.
                             // So just provide the Future trait.
                             let impl_bound = Binders::empty(
-                                &Interner,
+                                Interner,
                                 WhereClause::Implemented(TraitRef {
                                     trait_id: to_chalk_trait_id(future_trait),
-                                    substitution: Substitution::empty(&Interner),
+                                    substitution: Substitution::empty(Interner),
                                 }),
                             );
                             Some(vec![impl_bound])
@@ -206,7 +206,14 @@ impl TyExt for Ty {
                             None
                         }
                     }
-                    ImplTraitId::ReturnTypeImplTrait(..) => None,
+                    ImplTraitId::ReturnTypeImplTrait(func, idx) => {
+                        db.return_type_impl_traits(func).map(|it| {
+                            let data = (*it)
+                                .as_ref()
+                                .map(|rpit| rpit.impl_traits[idx as usize].bounds.clone());
+                            data.substitute(Interner, &subst).into_value_and_skipped_binders().0
+                        })
+                    }
                 }
             }
             TyKind::Alias(AliasTy::Opaque(opaque_ty)) => {
@@ -217,7 +224,7 @@ impl TyExt for Ty {
                             let data = (*it)
                                 .as_ref()
                                 .map(|rpit| rpit.impl_traits[idx as usize].bounds.clone());
-                            data.substitute(&Interner, &opaque_ty.substitution)
+                            data.substitute(Interner, &opaque_ty.substitution)
                         })
                     }
                     // It always has an parameter for Future::Output type.
@@ -235,16 +242,16 @@ impl TyExt for Ty {
                         let substs = TyBuilder::type_params_subst(db, id.parent);
                         let predicates = db
                             .generic_predicates(id.parent)
-                            .into_iter()
-                            .map(|pred| pred.clone().substitute(&Interner, &substs))
+                            .iter()
+                            .map(|pred| pred.clone().substitute(Interner, &substs))
                             .filter(|wc| match &wc.skip_binders() {
                                 WhereClause::Implemented(tr) => {
-                                    &tr.self_type_parameter(&Interner) == self
+                                    &tr.self_type_parameter(Interner) == self
                                 }
                                 WhereClause::AliasEq(AliasEq {
                                     alias: AliasTy::Projection(proj),
                                     ty: _,
-                                }) => &proj.self_type_parameter(&Interner) == self,
+                                }) => &proj.self_type_parameter(Interner) == self,
                                 _ => false,
                             })
                             .collect::<Vec<_>>();
@@ -259,10 +266,10 @@ impl TyExt for Ty {
     }
 
     fn associated_type_parent_trait(&self, db: &dyn HirDatabase) -> Option<TraitId> {
-        match self.kind(&Interner) {
+        match self.kind(Interner) {
             TyKind::AssociatedType(id, ..) => {
                 match from_assoc_type_id(*id).lookup(db.upcast()).container {
-                    AssocContainerId::TraitId(trait_id) => Some(trait_id),
+                    ItemContainerId::TraitId(trait_id) => Some(trait_id),
                     _ => None,
                 }
             }
@@ -271,7 +278,7 @@ impl TyExt for Ty {
                     .lookup(db.upcast())
                     .container
                 {
-                    AssocContainerId::TraitId(trait_id) => Some(trait_id),
+                    ItemContainerId::TraitId(trait_id) => Some(trait_id),
                     _ => None,
                 }
             }
@@ -280,7 +287,7 @@ impl TyExt for Ty {
     }
 
     fn equals_ctor(&self, other: &Ty) -> bool {
-        match (self.kind(&Interner), other.kind(&Interner)) {
+        match (self.kind(Interner), other.kind(Interner)) {
             (TyKind::Adt(adt, ..), TyKind::Adt(adt2, ..)) => adt == adt2,
             (TyKind::Slice(_), TyKind::Slice(_)) | (TyKind::Array(_, _), TyKind::Array(_, _)) => {
                 true
@@ -325,7 +332,7 @@ impl ProjectionTyExt for ProjectionTy {
 
     fn trait_(&self, db: &dyn HirDatabase) -> TraitId {
         match from_assoc_type_id(self.associated_ty_id).lookup(db.upcast()).container {
-            AssocContainerId::TraitId(it) => it,
+            ItemContainerId::TraitId(it) => it,
             _ => panic!("projection ty without parent trait"),
         }
     }
