@@ -1,12 +1,9 @@
 use std::{iter::once, mem};
 
 use hir::Semantics;
-use ide_db::{base_db::FileRange, RootDatabase};
+use ide_db::{base_db::FileRange, helpers::pick_best_token, RootDatabase};
 use itertools::Itertools;
-use syntax::{
-    algo, ast, match_ast, AstNode, SyntaxElement, SyntaxKind, SyntaxNode, SyntaxToken, TextRange,
-    TokenAtOffset,
-};
+use syntax::{algo, ast, match_ast, AstNode, SyntaxElement, SyntaxKind, SyntaxNode, TextRange};
 use text_edit::{TextEdit, TextEditBuilder};
 
 #[derive(Copy, Clone, Debug)]
@@ -36,7 +33,14 @@ pub(crate) fn move_item(
     let file = sema.parse(range.file_id);
 
     let item = if range.range.is_empty() {
-        SyntaxElement::Token(pick_best(file.syntax().token_at_offset(range.range.start()))?)
+        SyntaxElement::Token(pick_best_token(
+            file.syntax().token_at_offset(range.range.start()),
+            |kind| match kind {
+                SyntaxKind::IDENT | SyntaxKind::LIFETIME_IDENT => 2,
+                kind if kind.is_trivia() => 0,
+                _ => 1,
+            },
+        )?)
     } else {
         file.syntax().covering_element(range.range)
     };
@@ -116,13 +120,10 @@ fn swap_sibling_in_list<A: AstNode + Clone, I: Iterator<Item = A>>(
     range: TextRange,
     direction: Direction,
 ) -> Option<TextEdit> {
-    let list_lookup = list
-        .tuple_windows()
-        .filter(|(l, r)| match direction {
-            Direction::Up => r.syntax().text_range().contains_range(range),
-            Direction::Down => l.syntax().text_range().contains_range(range),
-        })
-        .next();
+    let list_lookup = list.tuple_windows().find(|(l, r)| match direction {
+        Direction::Up => r.syntax().text_range().contains_range(range),
+        Direction::Down => l.syntax().text_range().contains_range(range),
+    });
 
     if let Some((l, r)) = list_lookup {
         Some(replace_nodes(range, l.syntax(), r.syntax()))
@@ -170,18 +171,6 @@ fn replace_nodes<'a>(
     edit.finish()
 }
 
-fn pick_best(tokens: TokenAtOffset<SyntaxToken>) -> Option<SyntaxToken> {
-    return tokens.max_by_key(priority);
-
-    fn priority(n: &SyntaxToken) -> usize {
-        match n.kind() {
-            SyntaxKind::IDENT | SyntaxKind::LIFETIME_IDENT => 2,
-            kind if kind.is_trivia() => 0,
-            _ => 1,
-        }
-    }
-}
-
 #[cfg(test)]
 mod tests {
     use crate::fixture;
@@ -211,18 +200,18 @@ fn main() {
         }
     };
 }
-            "#,
+"#,
             expect![[r#"
-fn main() {
-    match true {
-        false =>$0 {
-            println!("Test");
-        },
-        true => {
-            println!("Hello, world");
-        }
-    };
-}
+                fn main() {
+                    match true {
+                        false =>$0 {
+                            println!("Test");
+                        }
+                        true => {
+                            println!("Hello, world");
+                        },
+                    };
+                }
             "#]],
             Direction::Up,
         );
@@ -242,18 +231,18 @@ fn main() {
         }
     };
 }
-            "#,
+"#,
             expect![[r#"
-fn main() {
-    match true {
-        false => {
-            println!("Test");
-        },
-        true =>$0 {
-            println!("Hello, world");
-        }
-    };
-}
+                fn main() {
+                    match true {
+                        false => {
+                            println!("Test");
+                        }
+                        true =>$0 {
+                            println!("Hello, world");
+                        },
+                    };
+                }
             "#]],
             Direction::Down,
         );
@@ -273,18 +262,18 @@ fn main() {
         }
     };
 }
-            "#,
+"#,
             expect![[r#"
-fn main() {
-    match true {
-        true => {
-            println!("Hello, world");
-        },
-        false => {
-            println!("Test");
-        }
-    };
-}
+                fn main() {
+                    match true {
+                        true => {
+                            println!("Hello, world");
+                        },
+                        false => {
+                            println!("Test");
+                        }
+                    };
+                }
             "#]],
             Direction::Up,
         );
@@ -298,12 +287,12 @@ fn main() {
     let test = 123;
     let test2$0$0 = 456;
 }
-            "#,
+"#,
             expect![[r#"
-fn main() {
-    let test2$0 = 456;
-    let test = 123;
-}
+                fn main() {
+                    let test2$0 = 456;
+                    let test = 123;
+                }
             "#]],
             Direction::Up,
         );
@@ -317,12 +306,12 @@ fn main() {
     println!("Hello, world");
     println!("All I want to say is...");$0$0
 }
-            "#,
+"#,
             expect![[r#"
-fn main() {
-    println!("All I want to say is...");$0
-    println!("Hello, world");
-}
+                fn main() {
+                    println!("All I want to say is...");$0
+                    println!("Hello, world");
+                }
             "#]],
             Direction::Up,
         );
@@ -335,15 +324,15 @@ fn main() {
         println!("Test");
     }$0$0
 }
-            "#,
+"#,
             expect![[r#"
-fn main() {
-    if true {
-        println!("Test");
-    }$0
+                fn main() {
+                    if true {
+                        println!("Test");
+                    }$0
 
-    println!("Hello, world");
-}
+                    println!("Hello, world");
+                }
             "#]],
             Direction::Up,
         );
@@ -356,15 +345,15 @@ fn main() {
         println!("Test");
     }$0$0
 }
-            "#,
+"#,
             expect![[r#"
-fn main() {
-    for i in 0..10 {
-        println!("Test");
-    }$0
+                fn main() {
+                    for i in 0..10 {
+                        println!("Test");
+                    }$0
 
-    println!("Hello, world");
-}
+                    println!("Hello, world");
+                }
             "#]],
             Direction::Up,
         );
@@ -377,15 +366,15 @@ fn main() {
         println!("Test");
     }$0$0
 }
-            "#,
+"#,
             expect![[r#"
-fn main() {
-    loop {
-        println!("Test");
-    }$0
+                fn main() {
+                    loop {
+                        println!("Test");
+                    }$0
 
-    println!("Hello, world");
-}
+                    println!("Hello, world");
+                }
             "#]],
             Direction::Up,
         );
@@ -398,15 +387,15 @@ fn main() {
         println!("Test");
     }$0$0
 }
-            "#,
+"#,
             expect![[r#"
-fn main() {
-    while true {
-        println!("Test");
-    }$0
+                fn main() {
+                    while true {
+                        println!("Test");
+                    }$0
 
-    println!("Hello, world");
-}
+                    println!("Hello, world");
+                }
             "#]],
             Direction::Up,
         );
@@ -417,13 +406,13 @@ fn main() {
 
     return 123;$0$0
 }
-            "#,
+"#,
             expect![[r#"
-fn main() {
-    return 123;$0
+                fn main() {
+                    return 123;$0
 
-    println!("Hello, world");
-}
+                    println!("Hello, world");
+                }
             "#]],
             Direction::Up,
         );
@@ -437,12 +426,12 @@ fn main() {
     println!("All I want to say is...");$0$0
     println!("Hello, world");
 }
-            "#,
+"#,
             expect![[r#"
-fn main() {
-    println!("All I want to say is...");
-    println!("Hello, world");
-}
+                fn main() {
+                    println!("All I want to say is...");
+                    println!("Hello, world");
+                }
             "#]],
             Direction::Up,
         );
@@ -455,11 +444,11 @@ fn main() {
 fn main() {}
 
 fn foo() {}$0$0
-            "#,
+"#,
             expect![[r#"
-fn foo() {}$0
+                fn foo() {}$0
 
-fn main() {}
+                fn main() {}
             "#]],
             Direction::Up,
         );
@@ -474,13 +463,13 @@ struct Yay;
 trait Wow {}
 
 impl Wow for Yay $0$0{}
-            "#,
+"#,
             expect![[r#"
-struct Yay;
+                struct Yay;
 
-impl Wow for Yay $0{}
+                impl Wow for Yay $0{}
 
-trait Wow {}
+                trait Wow {}
             "#]],
             Direction::Up,
         );
@@ -492,10 +481,10 @@ trait Wow {}
             r#"
 use std::vec::Vec;
 use std::collections::HashMap$0$0;
-            "#,
+"#,
             expect![[r#"
-use std::collections::HashMap$0;
-use std::vec::Vec;
+                use std::collections::HashMap$0;
+                use std::vec::Vec;
             "#]],
             Direction::Up,
         );
@@ -513,16 +502,16 @@ fn main() {
         _ => {}
     };$0
 }
-            "#,
+"#,
             expect![[r#"
-fn main() {
-    match test {
-        456 => {},
-        _ => {}
-    };
+                fn main() {
+                    match test {
+                        456 => {},
+                        _ => {}
+                    };
 
-    let test = 123;
-}
+                    let test = 123;
+                }
             "#]],
             Direction::Up,
         );
@@ -537,22 +526,22 @@ fn test(one: i32, two$0$0: u32) {}
 fn main() {
     test(123, 456);
 }
-            "#,
+"#,
             expect![[r#"
-fn test(two$0: u32, one: i32) {}
+                fn test(two$0: u32, one: i32) {}
 
-fn main() {
-    test(123, 456);
-}
+                fn main() {
+                    test(123, 456);
+                }
             "#]],
             Direction::Up,
         );
         check(
             r#"
 fn f($0$0arg: u8, arg2: u16) {}
-            "#,
+"#,
             expect![[r#"
-fn f(arg2: u16, $0arg: u8) {}
+                fn f(arg2: u16, $0arg: u8) {}
             "#]],
             Direction::Down,
         );
@@ -567,13 +556,13 @@ fn test(one: i32, two: u32) {}
 fn main() {
     test(123, 456$0$0);
 }
-            "#,
+"#,
             expect![[r#"
-fn test(one: i32, two: u32) {}
+                fn test(one: i32, two: u32) {}
 
-fn main() {
-    test(456$0, 123);
-}
+                fn main() {
+                    test(456$0, 123);
+                }
             "#]],
             Direction::Up,
         );
@@ -588,13 +577,13 @@ fn test(one: i32, two: u32) {}
 fn main() {
     test(123$0$0, 456);
 }
-            "#,
+"#,
             expect![[r#"
-fn test(one: i32, two: u32) {}
+                fn test(one: i32, two: u32) {}
 
-fn main() {
-    test(456, 123$0);
-}
+                fn main() {
+                    test(456, 123$0);
+                }
             "#]],
             Direction::Down,
         );
@@ -609,13 +598,13 @@ fn test(one: i32, two: u32) {}
 fn main() {
     test(123$0$0, 456);
 }
-            "#,
+"#,
             expect![[r#"
-fn test(one: i32, two: u32) {}
+                fn test(one: i32, two: u32) {}
 
-fn main() {
-    test(123, 456);
-}
+                fn main() {
+                    test(123, 456);
+                }
             "#]],
             Direction::Up,
         );
@@ -628,11 +617,11 @@ fn main() {
 struct Test<A, B$0$0>(A, B);
 
 fn main() {}
-            "#,
+"#,
             expect![[r#"
-struct Test<B$0, A>(A, B);
+                struct Test<B$0, A>(A, B);
 
-fn main() {}
+                fn main() {}
             "#]],
             Direction::Up,
         );
@@ -647,13 +636,13 @@ struct Test<A, B>(A, B);
 fn main() {
     let t = Test::<i32, &str$0$0>(123, "yay");
 }
-            "#,
+"#,
             expect![[r#"
-struct Test<A, B>(A, B);
+                struct Test<A, B>(A, B);
 
-fn main() {
-    let t = Test::<&str$0, i32>(123, "yay");
-}
+                fn main() {
+                    let t = Test::<&str$0, i32>(123, "yay");
+                }
             "#]],
             Direction::Up,
         );
@@ -669,14 +658,14 @@ enum Hello {
 }
 
 fn main() {}
-            "#,
+"#,
             expect![[r#"
-enum Hello {
-    Two$0,
-    One
-}
+                enum Hello {
+                    Two$0,
+                    One
+                }
 
-fn main() {}
+                fn main() {}
             "#]],
             Direction::Up,
         );
@@ -693,15 +682,15 @@ trait Two {}
 fn test<T: One + Two$0$0>(t: T) {}
 
 fn main() {}
-            "#,
+"#,
             expect![[r#"
-trait One {}
+                trait One {}
 
-trait Two {}
+                trait Two {}
 
-fn test<T: Two$0 + One>(t: T) {}
+                fn test<T: Two$0 + One>(t: T) {}
 
-fn main() {}
+                fn main() {}
             "#]],
             Direction::Up,
         );
@@ -730,27 +719,27 @@ impl Yay for Test {
         println!("Mmmm");
     }
 }
-            "#,
+"#,
             expect![[r#"
-struct Test;
+                struct Test;
 
-trait Yay {
-    type One;
+                trait Yay {
+                    type One;
 
-    type Two;
+                    type Two;
 
-    fn inner();
-}
+                    fn inner();
+                }
 
-impl Yay for Test {
-    type One = i32;
+                impl Yay for Test {
+                    type One = i32;
 
-    fn inner() {$0
-        println!("Mmmm");
-    }
+                    fn inner() {$0
+                        println!("Mmmm");
+                    }
 
-    type Two = u32;
-}
+                    type Two = u32;
+                }
             "#]],
             Direction::Up,
         );
@@ -769,17 +758,17 @@ fn test() {
         fn inner() {}
     }
 }
-            "#,
+"#,
             expect![[r#"
-fn test() {
-    mod hi {$0
-        fn inner() {}
-    }
+                fn test() {
+                    mod hi {$0
+                        fn inner() {}
+                    }
 
-    mod hello {
-        fn inner() {}
-    }
-}
+                    mod hello {
+                        fn inner() {}
+                    }
+                }
             "#]],
             Direction::Up,
         );
@@ -796,16 +785,16 @@ enum FooBar {
 }
 
 fn main() {}
-            "#,
-            expect![[r#"
-fn main() {}
+"#,
+            expect![[r##"
+                fn main() {}
 
-$0#[derive(Debug)]
-enum FooBar {
-    Foo,
-    Bar,
-}
-            "#]],
+                $0#[derive(Debug)]
+                enum FooBar {
+                    Foo,
+                    Bar,
+                }
+            "##]],
             Direction::Down,
         );
         check(
@@ -816,14 +805,14 @@ $0$0enum FooBar {
 }
 
 fn main() {}
-            "#,
+"#,
             expect![[r#"
-fn main() {}
+                fn main() {}
 
-$0enum FooBar {
-    Foo,
-    Bar,
-}
+                $0enum FooBar {
+                    Foo,
+                    Bar,
+                }
             "#]],
             Direction::Down,
         );
@@ -836,15 +825,15 @@ trait SomeTrait {}
 $0$0impl SomeTrait for Test {}
 
 fn main() {}
-            "#,
+"#,
             expect![[r#"
-struct Test;
+                struct Test;
 
-$0impl SomeTrait for Test {}
+                $0impl SomeTrait for Test {}
 
-trait SomeTrait {}
+                trait SomeTrait {}
 
-fn main() {}
+                fn main() {}
             "#]],
             Direction::Up,
         );
@@ -860,14 +849,14 @@ enum FooBar {
 }$0$0
 
 fn main() {}
-            "#,
+"#,
             expect![[r#"
-fn main() {}
+                fn main() {}
 
-enum FooBar {
-    Foo,
-    Bar,
-}$0
+                enum FooBar {
+                    Foo,
+                    Bar,
+                }$0
             "#]],
             Direction::Down,
         );
@@ -880,15 +869,15 @@ trait SomeTrait {}
 impl SomeTrait for Test {}$0$0
 
 fn main() {}
-            "#,
+"#,
             expect![[r#"
-struct Test;
+                struct Test;
 
-impl SomeTrait for Test {}$0
+                impl SomeTrait for Test {}$0
 
-trait SomeTrait {}
+                trait SomeTrait {}
 
-fn main() {}
+                fn main() {}
             "#]],
             Direction::Up,
         );
