@@ -11,7 +11,7 @@ use ide_db::{
 };
 use itertools::Itertools;
 use profile::{memory_usage, Bytes};
-use rustc_hash::FxHashMap;
+use std::env;
 use stdx::format_to;
 use syntax::{ast, Parse, SyntaxNode};
 
@@ -31,17 +31,20 @@ fn macro_syntax_tree_stats(db: &RootDatabase) -> SyntaxTreeStats {
 //
 // | VS Code | **Rust Analyzer: Status**
 // |===
+// image::https://user-images.githubusercontent.com/48062697/113065584-05f34500-91b1-11eb-98cc-5c196f76be7f.gif[]
 pub(crate) fn status(db: &RootDatabase, file_id: Option<FileId>) -> String {
     let mut buf = String::new();
     format_to!(buf, "{}\n", FileTextQuery.in_db(db).entries::<FilesStats>());
     format_to!(buf, "{}\n", LibrarySymbolsQuery.in_db(db).entries::<LibrarySymbolsStats>());
     format_to!(buf, "{}\n", syntax_tree_stats(db));
-    format_to!(buf, "{} (macros)\n", macro_syntax_tree_stats(db));
-    format_to!(buf, "{} total\n", memory_usage());
-    format_to!(buf, "\ncounts:\n{}", profile::countme::get_all());
+    format_to!(buf, "{} (Macros)\n", macro_syntax_tree_stats(db));
+    format_to!(buf, "{} in total\n", memory_usage());
+    if env::var("RA_COUNT").is_ok() {
+        format_to!(buf, "\nCounts:\n{}", profile::countme::get_all());
+    }
 
     if let Some(file_id) = file_id {
-        format_to!(buf, "\nfile info:\n");
+        format_to!(buf, "\nFile info:\n");
         let krate = crate::parent_module::crate_for(db, file_id).pop();
         match krate {
             Some(krate) => {
@@ -50,19 +53,19 @@ pub(crate) fn status(db: &RootDatabase, file_id: Option<FileId>) -> String {
                     Some(it) => format!("{}({:?})", it, krate),
                     None => format!("{:?}", krate),
                 };
-                format_to!(buf, "crate: {}\n", display_crate(krate));
+                format_to!(buf, "Crate: {}\n", display_crate(krate));
                 let deps = crate_graph[krate]
                     .dependencies
                     .iter()
                     .map(|dep| format!("{}={:?}", dep.name, dep.crate_id))
                     .format(", ");
-                format_to!(buf, "deps: {}\n", deps);
+                format_to!(buf, "Dependencies: {}\n", deps);
             }
-            None => format_to!(buf, "does not belong to any crate"),
+            None => format_to!(buf, "Does not belong to any crate"),
         }
     }
 
-    buf
+    buf.trim().to_string()
 }
 
 #[derive(Default)]
@@ -73,7 +76,7 @@ struct FilesStats {
 
 impl fmt::Display for FilesStats {
     fn fmt(&self, fmt: &mut fmt::Formatter) -> fmt::Result {
-        write!(fmt, "{} ({}) files", self.total, self.size)
+        write!(fmt, "{} of files", self.size)
     }
 }
 
@@ -99,7 +102,7 @@ pub(crate) struct SyntaxTreeStats {
 
 impl fmt::Display for SyntaxTreeStats {
     fn fmt(&self, fmt: &mut fmt::Formatter) -> fmt::Result {
-        write!(fmt, "{} trees, {} retained", self.total, self.retained)
+        write!(fmt, "{} trees, {} preserved", self.total, self.retained)
     }
 }
 
@@ -141,24 +144,20 @@ struct LibrarySymbolsStats {
 
 impl fmt::Display for LibrarySymbolsStats {
     fn fmt(&self, fmt: &mut fmt::Formatter) -> fmt::Result {
-        write!(fmt, "{} ({}) index symbols", self.total, self.size)
+        write!(fmt, "{} of index symbols ({})", self.size, self.total)
     }
 }
 
-impl FromIterator<TableEntry<(), Arc<FxHashMap<SourceRootId, SymbolIndex>>>>
-    for LibrarySymbolsStats
-{
+impl FromIterator<TableEntry<SourceRootId, Arc<SymbolIndex>>> for LibrarySymbolsStats {
     fn from_iter<T>(iter: T) -> LibrarySymbolsStats
     where
-        T: IntoIterator<Item = TableEntry<(), Arc<FxHashMap<SourceRootId, SymbolIndex>>>>,
+        T: IntoIterator<Item = TableEntry<SourceRootId, Arc<SymbolIndex>>>,
     {
         let mut res = LibrarySymbolsStats::default();
         for entry in iter {
-            let value = entry.value.unwrap();
-            for symbols in value.values() {
-                res.total += symbols.len();
-                res.size += symbols.memory_size();
-            }
+            let symbols = entry.value.unwrap();
+            res.total += symbols.len();
+            res.size += symbols.memory_size();
         }
         res
     }
