@@ -14,7 +14,7 @@ use crate::{AssistContext, AssistId, AssistKind, Assists};
 
 // Assist: replace_let_with_if_let
 //
-// Replaces `let` with an `if-let`.
+// Replaces `let` with an `if let`.
 //
 // ```
 // # enum Option<T> { Some(T), None }
@@ -38,18 +38,20 @@ use crate::{AssistContext, AssistId, AssistKind, Assists};
 // ```
 pub(crate) fn replace_let_with_if_let(acc: &mut Assists, ctx: &AssistContext) -> Option<()> {
     let let_kw = ctx.find_token_syntax_at_offset(T![let])?;
-    let let_stmt = let_kw.ancestors().find_map(ast::LetStmt::cast)?;
+    let let_stmt = let_kw.parent().and_then(ast::LetStmt::cast)?;
     let init = let_stmt.initializer()?;
     let original_pat = let_stmt.pat()?;
-    let ty = ctx.sema.type_of_expr(&init)?;
-    let happy_variant = TryEnum::from_ty(&ctx.sema, &ty).map(|it| it.happy_case());
 
     let target = let_kw.text_range();
     acc.add(
         AssistId("replace_let_with_if_let", AssistKind::RefactorRewrite),
-        "Replace with if-let",
+        "Replace let with if let",
         target,
         |edit| {
+            let ty = ctx.sema.type_of_expr(&init);
+            let happy_variant = ty
+                .and_then(|ty| TryEnum::from_ty(&ctx.sema, &ty.adjusted()))
+                .map(|it| it.happy_case());
             let pat = match happy_variant {
                 None => original_pat,
                 Some(var_name) => {
