@@ -1,12 +1,9 @@
 //! This modules defines type to represent changes to the source code, that flow
 //! from the server to the client.
 //!
-//! It can be viewed as a dual for `AnalysisChange`.
+//! It can be viewed as a dual for `Change`.
 
-use std::{
-    collections::hash_map::Entry,
-    iter::{self, FromIterator},
-};
+use std::{collections::hash_map::Entry, iter};
 
 use base_db::{AnchoredPathBuf, FileId};
 use rustc_hash::FxHashMap;
@@ -32,11 +29,13 @@ impl SourceChange {
 
     pub fn from_text_edit(file_id: FileId, edit: TextEdit) -> Self {
         SourceChange {
-            source_file_edits: FxHashMap::from_iter(iter::once((file_id, edit))),
+            source_file_edits: iter::once((file_id, edit)).collect(),
             ..Default::default()
         }
     }
 
+    /// Inserts a [`TextEdit`] for the given [`FileId`]. This properly handles merging existing
+    /// edits for a file if some already exist.
     pub fn insert_source_edit(&mut self, file_id: FileId, edit: TextEdit) {
         match self.source_file_edits.entry(file_id) {
             Entry::Occupied(mut entry) => {
@@ -55,6 +54,13 @@ impl SourceChange {
     pub fn get_source_edit(&self, file_id: FileId) -> Option<&TextEdit> {
         self.source_file_edits.get(&file_id)
     }
+
+    pub fn merge(mut self, other: SourceChange) -> SourceChange {
+        self.extend(other.source_file_edits);
+        self.extend(other.file_system_edits);
+        self.is_snippet |= other.is_snippet;
+        self
+    }
 }
 
 impl Extend<(FileId, TextEdit)> for SourceChange {
@@ -63,6 +69,12 @@ impl Extend<(FileId, TextEdit)> for SourceChange {
     }
 }
 
+impl Extend<FileSystemEdit> for SourceChange {
+    fn extend<T: IntoIterator<Item = FileSystemEdit>>(&mut self, iter: T) {
+        iter.into_iter().for_each(|edit| self.push_file_system_edit(edit));
+    }
+}
+
 impl From<FxHashMap<FileId, TextEdit>> for SourceChange {
     fn from(source_file_edits: FxHashMap<FileId, TextEdit>) -> SourceChange {
         SourceChange { source_file_edits, file_system_edits: Vec::new(), is_snippet: false }
