@@ -1,60 +1,39 @@
-//! FIXME: write short doc here
-
 mod atom;
 
+use super::*;
+
 pub(crate) use self::atom::{block_expr, match_arm_list};
 pub(super) use self::atom::{literal, LITERAL_FIRST};
-use super::*;
 
-pub(super) enum StmtWithSemi {
-    Yes,
-    No,
+#[derive(PartialEq, Eq)]
+pub(super) enum Semicolon {
+    Required,
     Optional,
+    Forbidden,
 }
 
 const EXPR_FIRST: TokenSet = LHS_FIRST;
 
-pub(super) fn expr(p: &mut Parser) -> (Option<CompletedMarker>, BlockLike) {
+pub(super) fn expr(p: &mut Parser) -> bool {
     let r = Restrictions { forbid_structs: false, prefer_stmt: false };
-    expr_bp(p, r, 1)
+    expr_bp(p, None, r, 1).is_some()
 }
 
-pub(super) fn expr_with_attrs(p: &mut Parser) -> bool {
-    let m = p.start();
-    let has_attrs = p.at(T![#]);
-    attributes::outer_attrs(p);
-
-    let (cm, _block_like) = expr(p);
-    let success = cm.is_some();
-
-    match (has_attrs, cm) {
-        (true, Some(cm)) => {
-            let kind = cm.kind();
-            cm.undo_completion(p).abandon(p);
-            m.complete(p, kind);
-        }
-        _ => m.abandon(p),
-    }
-
-    success
-}
-
-pub(super) fn expr_stmt(p: &mut Parser) -> (Option<CompletedMarker>, BlockLike) {
+pub(super) fn expr_stmt(p: &mut Parser, m: Option<Marker>) -> Option<(CompletedMarker, BlockLike)> {
     let r = Restrictions { forbid_structs: false, prefer_stmt: true };
-    expr_bp(p, r, 1)
+    expr_bp(p, m, r, 1)
 }
 
 fn expr_no_struct(p: &mut Parser) {
     let r = Restrictions { forbid_structs: true, prefer_stmt: false };
-    expr_bp(p, r, 1);
+    expr_bp(p, None, r, 1);
 }
 
-fn is_expr_stmt_attr_allowed(kind: SyntaxKind) -> bool {
-    let forbid = matches!(kind, BIN_EXPR | RANGE_EXPR);
-    !forbid
-}
+pub(super) fn stmt(p: &mut Parser, semicolon: Semicolon) {
+    if p.eat(T![;]) {
+        return;
+    }
 
-pub(super) fn stmt(p: &mut Parser, with_semi: StmtWithSemi) {
     let m = p.start();
     // test attr_on_expr_stmt
     // fn foo() {
@@ -63,110 +42,87 @@ pub(super) fn stmt(p: &mut Parser, with_semi: StmtWithSemi) {
     //     #[C] #[D] {}
     //     #[D] return ();
     // }
-    let has_attrs = p.at(T![#]);
     attributes::outer_attrs(p);
 
     if p.at(T![let]) {
-        let_stmt(p, m, with_semi);
+        let_stmt(p, m, semicolon);
         return;
     }
 
     // test block_items
     // fn a() { fn b() {} }
-    let m = match items::maybe_item(p, m) {
+    let m = match items::opt_item(p, m) {
         Ok(()) => return,
         Err(m) => m,
     };
 
-    let (cm, blocklike) = expr_stmt(p);
-    let kind = cm.as_ref().map(|cm| cm.kind()).unwrap_or(ERROR);
-
-    if has_attrs && !is_expr_stmt_attr_allowed(kind) {
-        // test_err attr_on_expr_not_allowed
-        // fn foo() {
-        //    #[A] 1 + 2;
-        //    #[B] if true {};
-        // }
-        p.error(format!("attributes are not allowed on {:?}", kind));
-    }
-
-    if p.at(T!['}']) {
-        // test attr_on_last_expr_in_block
-        // fn foo() {
-        //     { #[A] bar!()? }
-        //     #[B] &()
-        // }
-        if let Some(cm) = cm {
-            cm.undo_completion(p).abandon(p);
-            m.complete(p, kind);
-        } else {
-            m.abandon(p);
-        }
-    } else {
-        // test no_semi_after_block
-        // fn foo() {
-        //     if true {}
-        //     loop {}
-        //     match () {}
-        //     while true {}
-        //     for _ in () {}
-        //     {}
-        //     {}
-        //     macro_rules! test {
-        //          () => {}
-        //     }
-        //     test!{}
-        // }
-
-        match with_semi {
-            StmtWithSemi::Yes => {
-                if blocklike.is_block() {
-                    p.eat(T![;]);
-                } else {
-                    p.expect(T![;]);
+    if let Some((cm, blocklike)) = expr_stmt(p, Some(m)) {
+        if !(p.at(T!['}']) || (semicolon != Semicolon::Required && p.at(EOF))) {
+            // test no_semi_after_block
+            // fn foo() {
+            //     if true {}
+            //     loop {}
+            //     match () {}
+            //     while true {}
+            //     for _ in () {}
+            //     {}
+            //     {}
+            //     macro_rules! test {
+            //          () => {}
+            //     }
+            //     test!{}
+            // }
+            let m = cm.precede(p);
+            match semicolon {
+                Semicolon::Required => {
+                    if blocklike.is_block() {
+                        p.eat(T![;]);
+                    } else {
+                        p.expect(T![;]);
+                    }
                 }
-            }
-            StmtWithSemi::No => {}
-            StmtWithSemi::Optional => {
-                if p.at(T![;]) {
+                Semicolon::Optional => {
                     p.eat(T![;]);
                 }
+                Semicolon::Forbidden => (),
             }
+            m.complete(p, EXPR_STMT);
         }
-
-        m.complete(p, EXPR_STMT);
     }
 
     // test let_stmt
-    // fn foo() {
-    //     let a;
-    //     let b: i32;
-    //     let c = 92;
-    //     let d: i32 = 92;
-    //     let e: !;
-    //     let _: ! = {};
-    //     let f = #[attr]||{};
-    // }
-    fn let_stmt(p: &mut Parser, m: Marker, with_semi: StmtWithSemi) {
-        assert!(p.at(T![let]));
+    // fn f() { let x: i32 = 92; }
+    fn let_stmt(p: &mut Parser, m: Marker, with_semi: Semicolon) {
         p.bump(T![let]);
         patterns::pattern(p);
         if p.at(T![:]) {
+            // test let_stmt_ascription
+            // fn f() { let x: i32; }
             types::ascription(p);
         }
         if p.eat(T![=]) {
-            expressions::expr_with_attrs(p);
+            // test let_stmt_init
+            // fn f() { let x = 92; }
+            expressions::expr(p);
+        }
+
+        if p.at(T![else]) {
+            // test let_else
+            // fn f() { let Some(x) = opt else { return }; }
+
+            let m = p.start();
+            p.bump(T![else]);
+            block_expr(p);
+            m.complete(p, LET_ELSE);
         }
 
         match with_semi {
-            StmtWithSemi::Yes => {
-                p.expect(T![;]);
+            Semicolon::Forbidden => (),
+            Semicolon::Optional => {
+                p.eat(T![;]);
             }
-            StmtWithSemi::No => {}
-            StmtWithSemi::Optional => {
-                if p.at(T![;]) {
-                    p.eat(T![;]);
-                }
+            Semicolon::Required => {
+                p.expect(T![;]);
             }
         }
         m.complete(p, LET_STMT);
@@ -174,7 +130,6 @@ pub(super) fn stmt(p: &mut Parser, with_semi: StmtWithSemi) {
 }
 
 pub(super) fn expr_block_contents(p: &mut Parser) {
-    // This is checked by a validator
     attributes::inner_attrs(p);
 
     while !p.at(EOF) && !p.at(T!['}']) {
@@ -192,13 +147,7 @@ pub(super) fn expr_block_contents(p: &mut Parser) {
         //     fn f() {};
         //     struct S {};
         // }
-
-        if p.at(T![;]) {
-            p.bump(T![;]);
-            continue;
-        }
-
-        stmt(p, StmtWithSemi::Yes)
+        stmt(p, Semicolon::Required);
     }
 }
 
@@ -210,7 +159,7 @@ struct Restrictions {
 
 /// Binding powers of operators for a Pratt parser.
 ///
-/// See https://www.oilshell.org/blog/2016/11/03.html
+/// See <https://matklad.github.io/2020/04/13/simple-but-powerful-pratt-parsing.html>
 #[rustfmt::skip]
 fn current_op(p: &Parser) -> (u8, SyntaxKind) {
     const NOT_AN_OP: (u8, SyntaxKind) = (0, T![@]);
@@ -254,20 +203,34 @@ fn current_op(p: &Parser) -> (u8, SyntaxKind) {
 }
 
 // Parses expression with binding power of at least bp.
-fn expr_bp(p: &mut Parser, mut r: Restrictions, bp: u8) -> (Option<CompletedMarker>, BlockLike) {
+fn expr_bp(
+    p: &mut Parser,
+    m: Option<Marker>,
+    mut r: Restrictions,
+    bp: u8,
+) -> Option<(CompletedMarker, BlockLike)> {
+    let m = m.unwrap_or_else(|| {
+        let m = p.start();
+        attributes::outer_attrs(p);
+        m
+    });
     let mut lhs = match lhs(p, r) {
         Some((lhs, blocklike)) => {
-            // test stmt_bin_expr_ambiguity
-            // fn foo() {
-            //     let _ = {1} & 2;
-            //     {1} &2;
-            // }
+            let lhs = lhs.extend_to(p, m);
             if r.prefer_stmt && blocklike.is_block() {
-                return (Some(lhs), BlockLike::Block);
+                // test stmt_bin_expr_ambiguity
+                // fn f() {
+                //     let _ = {1} & 2;
+                //     {1} &2;
+                // }
+                return Some((lhs, BlockLike::Block));
             }
             lhs
         }
-        None => return (None, BlockLike::NotBlock),
+        None => {
+            m.abandon(p);
+            return None;
+        }
     };
 
     loop {
@@ -277,9 +240,7 @@ fn expr_bp(p: &mut Parser, mut r: Restrictions, bp: u8) -> (Option<CompletedMark
             break;
         }
         // test as_precedence
-        // fn foo() {
-        //     let _ = &1 as *const i32;
-        // }
+        // fn f() { let _ = &1 as *const i32; }
         if p.at(T![as]) {
             lhs = cast_expr(p, lhs);
             continue;
@@ -288,9 +249,7 @@ fn expr_bp(p: &mut Parser, mut r: Restrictions, bp: u8) -> (Option<CompletedMark
         p.bump(op);
 
         // test binop_resets_statementness
-        // fn foo() {
-        //     v = {1}&2;
-        // }
+        // fn f() { v = {1}&2; }
         r = Restrictions { prefer_stmt: false, ..r };
 
         if is_range {
@@ -309,14 +268,14 @@ fn expr_bp(p: &mut Parser, mut r: Restrictions, bp: u8) -> (Option<CompletedMark
             }
         }
 
-        expr_bp(p, Restrictions { prefer_stmt: false, ..r }, op_bp + 1);
+        expr_bp(p, None, Restrictions { prefer_stmt: false, ..r }, op_bp + 1);
         lhs = m.complete(p, if is_range { RANGE_EXPR } else { BIN_EXPR });
     }
-    (Some(lhs), BlockLike::NotBlock)
+    Some((lhs, BlockLike::NotBlock))
 }
 
 const LHS_FIRST: TokenSet =
-    atom::ATOM_EXPR_FIRST.union(token_set![T![&], T![*], T![!], T![.], T![-]]);
+    atom::ATOM_EXPR_FIRST.union(TokenSet::new(&[T![&], T![*], T![!], T![.], T![-]]));
 
 fn lhs(p: &mut Parser, r: Restrictions) -> Option<(CompletedMarker, BlockLike)> {
     let m;
@@ -335,10 +294,7 @@ fn lhs(p: &mut Parser, r: Restrictions) -> Option<(CompletedMarker, BlockLike)>
         T![&] => {
             m = p.start();
             p.bump(T![&]);
-            if p.at(IDENT)
-                && p.at_contextual_kw("raw")
-                && (p.nth_at(1, T![mut]) || p.nth_at(1, T![const]))
-            {
+            if p.at_contextual_kw(T![raw]) && (p.nth_at(1, T![mut]) || p.nth_at(1, T![const])) {
                 p.bump_remap(T![raw]);
                 p.bump_any();
             } else {
@@ -360,14 +316,15 @@ fn lhs(p: &mut Parser, r: Restrictions) -> Option<(CompletedMarker, BlockLike)>
         _ => {
             // test full_range_expr
             // fn foo() { xs[..]; }
-            for &op in [T![..=], T![..]].iter() {
+            for op in [T![..=], T![..]] {
                 if p.at(op) {
                     m = p.start();
                     p.bump(op);
                     if p.at_ts(EXPR_FIRST) && !(r.forbid_structs && p.at(T!['{'])) {
-                        expr_bp(p, r, 2);
+                        expr_bp(p, None, r, 2);
                     }
-                    return Some((m.complete(p, RANGE_EXPR), BlockLike::NotBlock));
+                    let cm = m.complete(p, RANGE_EXPR);
+                    return Some((cm, BlockLike::NotBlock));
                 }
             }
 
@@ -376,14 +333,16 @@ fn lhs(p: &mut Parser, r: Restrictions) -> Option<(CompletedMarker, BlockLike)>
             //    let mut p = F{x: 5};
             //    {p}.x = 10;
             // }
-            //
             let (lhs, blocklike) = atom::atom_expr(p, r)?;
-            return Some(postfix_expr(p, lhs, blocklike, !(r.prefer_stmt && blocklike.is_block())));
+            let (cm, block_like) =
+                postfix_expr(p, lhs, blocklike, !(r.prefer_stmt && blocklike.is_block()));
+            return Some((cm, block_like));
         }
     };
     // parse the interior of the unary expression
-    expr_bp(p, r, 255);
-    Some((m.complete(p, kind), BlockLike::NotBlock))
+    expr_bp(p, None, r, 255);
+    let cm = m.complete(p, kind);
+    Some((cm, BlockLike::NotBlock))
 }
 
 fn postfix_expr(
@@ -489,7 +448,7 @@ fn method_call_expr(p: &mut Parser, lhs: CompletedMarker) -> CompletedMarker {
     let m = lhs.precede(p);
     p.bump_any();
     name_ref(p);
-    type_args::opt_generic_arg_list(p, true);
+    generic_args::opt_generic_arg_list(p, true);
     if p.at(T!['(']) {
         arg_list(p);
     }
@@ -502,24 +461,17 @@ fn method_call_expr(p: &mut Parser, lhs: CompletedMarker) -> CompletedMarker {
 //     x.0.bar;
 //     x.0();
 // }
-
-// test_err bad_tuple_index_expr
-// fn foo() {
-//     x.0.;
-//     x.1i32;
-//     x.0x01;
-// }
 fn field_expr(p: &mut Parser, lhs: CompletedMarker) -> CompletedMarker {
     assert!(p.at(T![.]));
     let m = lhs.precede(p);
     p.bump(T![.]);
     if p.at(IDENT) || p.at(INT_NUMBER) {
-        name_ref_or_index(p)
+        name_ref_or_index(p);
     } else if p.at(FLOAT_NUMBER) {
         // FIXME: How to recover and instead parse INT + T![.]?
         p.bump_any();
     } else {
-        p.error("expected field name or number")
+        p.error("expected field name or number");
     }
     m.complete(p, FIELD_EXPR)
 }
@@ -561,7 +513,7 @@ fn arg_list(p: &mut Parser) {
         // fn main() {
         //     foo(#[attr] 92)
         // }
-        if !expr_with_attrs(p) {
+        if !expr(p) {
             break;
         }
         if !p.at(T![')']) && !p.expect(T![,]) {
